@@ -26,9 +26,9 @@
  * 
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23148 $
+ * $Revision: 23275 $
  * $Author: rjongbloed $
- * $Date: 2009-07-28 07:42:40 -0500 (Tue, 28 Jul 2009) $
+ * $Date: 2009-08-27 02:17:54 -0500 (Thu, 27 Aug 2009) $
  */
 
 #ifndef OPAL_LIDS_LIDEP_H
@@ -100,7 +100,7 @@ class OpalLineEndPoint : public OpalEndPoint
 
        The default behaviour is pure.
      */
-    virtual PBoolean MakeConnection(
+    virtual PSafePtr<OpalConnection> MakeConnection(
       OpalCall & call,          ///< Owner of connection
       const PString & party,    ///< Remote party to call
       void * userData = NULL,   ///< Arbitrary data to pass to connection
@@ -489,6 +489,15 @@ class OpalLineConnection : public OpalConnection
       */
     OpalLine & GetLine() { return line; }
 
+    /** Get the prompt tone used on POTS lines.
+        Defaults to OpalLineInterfaceDevice::DialTone.
+      */
+    OpalLineInterfaceDevice::CallProgressTones GetPromptTone() const { return m_promptTone; }
+
+    /** Set the prompt tone used on POTS lines.
+      */
+    void SetPromptTone(OpalLineInterfaceDevice::CallProgressTones tone) { m_promptTone = tone; }
+
     /** delay in msec to wait between the dial tone detection and dialing the dtmf 
       * @param uiDial the dial delay to set
      */
@@ -506,6 +515,7 @@ class OpalLineConnection : public OpalConnection
     bool              wasOffHook;
     unsigned          minimumRingCount;
     OpalLineInterfaceDevice::DialParams m_dialParams;
+    OpalLineInterfaceDevice::CallProgressTones m_promptTone;
 
     PDECLARE_NOTIFIER(PThread, OpalLineConnection, HandleIncoming);
     PThread         * handlerThread;
@@ -532,6 +542,9 @@ class OpalLineMediaStream : public OpalMediaStream
     );
   //@}
 
+    ~OpalLineMediaStream();
+
+
   /**@name Overrides of OpalMediaStream class */
   //@{
     /**Open the media stream.
@@ -599,16 +612,17 @@ class OpalLineMediaStream : public OpalMediaStream
     virtual PBoolean IsSynchronous() const;
 
     /**Indicate if the media stream requires a OpalMediaPatch thread (active patch).
-       This is called on the source stream and is passed the sink stream that the
-       patch will initially be using. The function could conditionally require
-       the patch thread to execute a thread reading and writing data, or prevent
-       it from doing so as it can do so in hardware in some way, e.g. if both
-       streams where on the same OpalLineInterfaceDevice.
+       This is called on the source/sink stream and is passed the sink/source
+       stream that the patch will initially be using. The function could
+       conditionally require the patch thread to execute a thread reading and
+       writing data, or prevent  it from doing so as it can do so in hardware
+       in some way.
 
-       The default behaviour simply returns true.
+       The default behaviour here determines if both streams are on the same
+       OpalLineInterfaceDevice and returns false if so.
       */
     virtual PBoolean RequiresPatchThread(
-      OpalMediaStream * sinkStream  ///< Sink stream for this source
+      OpalMediaStream * stream  ///< Other stream in patch
     ) const;
   //@}
 
