@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22096 $
+ * $Revision: 24038 $
  * $Author: rjongbloed $
- * $Date: 2009-02-21 03:25:23 +0000 (Sat, 21 Feb 2009) $
+ * $Date: 2010-02-09 22:21:59 -0600 (Tue, 09 Feb 2010) $
  */
 
 #ifndef OPAL_OPAL_MANAGER_H
@@ -38,13 +38,14 @@
 
 #include <opal/buildopts.h>
 
+#include <opal/pres_ent.h>
 #include <opal/call.h>
 #include <opal/connection.h> //OpalConnection::AnswerCallResponse
 #include <opal/guid.h>
-#include <opal/audiorecord.h>
 #include <codec/silencedetect.h>
 #include <codec/echocancel.h>
 #include <ptclib/pstun.h>
+#include <ptclib/url.h>
 
 #if OPAL_VIDEO
 #include <ptlib/videoio.h>
@@ -53,6 +54,7 @@
 class OpalEndPoint;
 class OpalMediaPatch;
 
+
 /**This class is the central manager for OPAL.
    The OpalManager embodies the root of the tree of objects that constitute an
    OPAL system. It contains all of the endpoints that make up the system. Other
@@ -146,6 +148,13 @@ class OpalManager : public PObject
        deleted) at any time due to the multithreaded nature of the OPAL
        system. 
      */
+    virtual PSafePtr<OpalCall> SetUpCall(
+      const PString & partyA,       ///<  The A party of call
+      const PString & partyB,       ///<  The B party of call
+      void * userData = NULL,       ///<  user data passed to Call and Connection
+      unsigned options = 0,         ///<  options passed to connection
+      OpalConnection::StringOptions * stringOptions = NULL   ///<  complex string options passed to call
+    );
     virtual PBoolean SetUpCall(
       const PString & partyA,       ///<  The A party of call
       const PString & partyB,       ///<  The B party of call
@@ -267,7 +276,6 @@ class OpalManager : public PObject
     virtual OpalCall * CreateCall(
       void * userData            ///<  user data passed to SetUpCall
     );
-    virtual OpalCall * CreateCall();
     OpalCall * InternalCreateCall();
 
     /**Destroy a call object.
@@ -282,10 +290,73 @@ class OpalManager : public PObject
       OpalCall * call
     );
 
-    /**Get next unique token ID for calls.
-       This is an internal function called by the OpalCall constructor.
+    /**Get next unique token ID for calls or connections.
+       This is an internal function called by the OpalCall and other
+       constructors.
       */
-    PString GetNextCallToken();
+    virtual PString GetNextToken(char prefix);
+  //@}
+
+  /**@name Presence & Instant Messaging management */
+  //@{
+    /**Add a presentity.
+       Returns a Read/Write locked pointer to presentity.
+      */
+    virtual PSafePtr<OpalPresentity> AddPresentity(
+      const PString & presentity  ///< Presentity URI
+    );
+
+    /**Get a presentity.
+      */
+    virtual PSafePtr<OpalPresentity> GetPresentity(
+      const PString & presentity,         ///< Presentity URI
+      PSafetyMode mode = PSafeReference   ///< Safety mode for presentity
+    );
+
+    /**Get all presentities.
+      */
+    virtual PStringList GetPresentities() const;
+
+    /**Remove a presentity.
+      */
+    virtual bool RemovePresentity(
+      const PString & presentity  ///< Presentity URI
+    );
+
+    /**Send text message
+     */
+    virtual PBoolean Message(
+      const PString & to, 
+      const PString & body
+    );
+    virtual PBoolean Message(
+      const PURL & to, 
+      const PString & type,
+      const PString & body,
+      PURL & from, 
+      PString & conversationId
+    );
+
+    /**Called when text message received
+     */
+    virtual void OnMessageReceived(
+      const PURL & from, 
+      const PString & fromName,
+      const PURL & to, 
+      const PString & type,
+      const PString & body,
+      const PString & conversationId
+    );
+
+#if OPAL_HAS_IM
+    /** Called when text message to be sent to remote end of a connection
+      */
+    virtual bool TransmitExternalIM(
+      OpalConnection & conn, 
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & frame
+    );
+#endif
   //@}
 
   /**@name Connection management */
@@ -324,7 +395,7 @@ class OpalManager : public PObject
 
        The default behaviour is pure.
      */
-    virtual PBoolean MakeConnection(
+    virtual PSafePtr<OpalConnection> MakeConnection(
       OpalCall & call,                   ///<  Owner of connection
       const PString & party,             ///<  Party to call
       void * userData = NULL,            ///<  user data to pass to connections
@@ -332,6 +403,15 @@ class OpalManager : public PObject
       OpalConnection::StringOptions * stringOptions = NULL
     );
 
+    /**Call back for a new connection has been constructed.
+       This is called after CreateConnection has returned a new connection.
+       It allows an application to make any custom adjustments to the
+       connection before it begins to process the protocol. behind it.
+      */
+    virtual void OnNewConnection(
+      OpalConnection & connection   ///< New connection just created
+    );
+
     /**Call back for answering an incoming call.
        This function is used for an application to control the answering of
        incoming calls.
@@ -363,13 +443,6 @@ class OpalManager : public PObject
       unsigned options,              ///<  options for new connection (can't use default as overrides will fail)
       OpalConnection::StringOptions * stringOptions
     );
-    virtual PBoolean OnIncomingConnection(
-      OpalConnection & connection,   ///<  Connection that is calling
-      unsigned options               ///<  options for new connection (can't use default as overrides will fail)
-    );
-    virtual PBoolean OnIncomingConnection(
-      OpalConnection & connection   ///<  Connection that is calling
-    );
 
     /**Route a connection to another connection from an endpoint.
 
@@ -425,9 +498,25 @@ class OpalManager : public PObject
       OpalConnection & connection   ///<  Connection that was established
     );
 
-    virtual OpalConnection::AnswerCallResponse
-       OnAnswerCall(OpalConnection & connection,
-                     const PString & caller
+    /**Call back for answering an incoming call.
+       This function is called after the connection has been acknowledged
+       but before the connection is established
+
+       This gives the application time to wait for some event before
+       signalling to the endpoint that the connection is to proceed. For
+       example the user pressing an "Answer call" button.
+
+       If AnswerCallDenied is returned the connection is aborted and the
+       connetion specific end call PDU is sent. If AnswerCallNow is returned 
+       then the connection proceeding, Finally if AnswerCallPending is returned then the
+       protocol negotiations are paused until the AnsweringCall() function is
+       called.
+
+       The default behaviour simply returns AnswerNow.
+     */
+    virtual OpalConnection::AnswerCallResponse OnAnswerCall(
+      OpalConnection & connection,    ///<  connection that is being answered
+       const PString & caller         ///<  caller
     );
 
     /**A call back function whenever a connection is "connected".
@@ -506,6 +595,18 @@ class OpalManager : public PObject
 
   /**@name Media Streams management */
   //@{
+    /**Get common media formats.
+       This is called by various places to get common media formats for the
+       basic connection classes.
+
+       The default behaviour uses the mediaFormatOrder and mediaFormatMask
+       member variables to adjust the mediaFormats list.
+      */
+    virtual OpalMediaFormatList GetCommonMediaFormats(
+      bool transportable,  ///< Include transportable media formats
+      bool pcmAudio        ///< Include raw PCM audio media formats
+    ) const;
+
     /**Adjust media formats available on a connection.
        This is called by a connection after it has called
        OpalCall::GetMediaFormats() to get all media formats that it can use so
@@ -516,6 +617,7 @@ class OpalManager : public PObject
        member variables to adjust the mediaFormats list.
       */
     virtual void AdjustMediaFormats(
+      bool local,                         ///<  Media formats a local ones to be presented to remote
       const OpalConnection & connection,  ///<  Connection that is about to use formats
       OpalMediaFormatList & mediaFormats  ///<  Media formats to use
     ) const;
@@ -548,7 +650,17 @@ class OpalManager : public PObject
       OpalMediaStream & stream    ///<  New media stream being opened
     );
 
-    /**Callback from the RTP session for statistics monitoring.
+    /**Create an RTP session.
+       This function allows an application to have the system create descendant
+       class versions of the RTP_UDP class. The application could use this to 
+	   modify the default RTP transfer behaviour.
+	   This is called when a connection determines that RTP is required for 
+	   transporting media.
+       The default behaviour returns a new RTP_UDP.
+      */
+	virtual RTP_UDP * CreateRTPSession (const RTP_Session::Params & params);
+
+	/**Callback from the RTP session for statistics monitoring.
        This is called every so many packets on the transmitter and receiver
        threads of the RTP session indicating that the statistics have been
        updated.
@@ -560,6 +672,68 @@ class OpalManager : public PObject
       const RTP_Session & session         ///<  Session with statistics
     );
 
+    /**Indicate is a local RTP connection.
+       This is called when a new media stream has been created and it has been
+       detected that media will be flowing between two RTP sessions within the
+       same process. An application could take advantage of this by optimising
+       the transfer in some way, rather than the full media path of codecs and
+       sockets which might not be necessary.
+
+       Note this is the complement to SetMediaBypass() as this function stops
+       RTP data from being sent/received, while SetMediaBypass() transfers
+       RTP data between the two endpoints.
+
+       The default behaviour returns false.
+
+       @return true if the application is going to execute some form of
+               bypass, and the media patch threads should not be started.
+      */
+    virtual bool OnLocalRTP(
+      OpalConnection & connection1, ///< First connection
+      OpalConnection & connection2, ///< Second connection
+      unsigned sessionID,           ///< Session ID of RTP session
+      bool opened                   ///< Media streams are opened/closed
+    ) const;
+
+    /**Set bypass mode for media.
+
+       Bypass the internal media handling, passing RTP data directly from
+       one call/connection to another.
+
+       This can be useful for back to back calls that happen to be the same
+       media format and you wish to avoid double decoding and encoding of
+       media. Note this scenario is not the same as two OpalConnections within
+       the same OpalCall, but two completely independent OpalCall where one
+       connection is to be bypassed. For example, two OpalCall instances might
+       have two SIPConnection instances and two OpalMixerConnection instances
+       connected via a single OpalMixerNode. Now while there are ONLY two
+       calls in the node, it is a waste to decode the audio, add to mixer and
+       re-encode it again. In practice this is identical to just bypassing the
+       mixer node completely, until a third party is added, then we need to
+       switch back to normal (non-bypassed) operation.
+
+       Note this is the complement to OnLocalRTP() as this function transfers
+       RTP data directly between the two endpoints, while OnLocalRTP() stops
+       the RTP data from being sent/received.
+
+       @return true if bypass is started/stopped, false if there was no such
+               call/connection/stream, the streams are incompatible formats
+               or a conflicting bypass is already in place.
+      */
+    bool SetMediaBypass(
+      const PString & token1, ///< First calls token
+      const PString & token2, ///< Second calls token
+      bool bypass,            ///< Bypass the media
+      unsigned sessionID = 0, ///< Session ID of media stream, 0 indicates all
+      bool network  = true    ///< Bypass the network connections of the calls
+    );
+    static bool SetMediaBypass(
+      OpalConnection & connection1, ///< First connection
+      OpalConnection & connection2, ///< Second connection
+      bool bypass,                  ///< Bypass the media
+      unsigned sessionID = 0        ///< Session ID of media stream, 0 indicates all
+    );
+
     /**Call back for closed a media stream.
 
        The default behaviour does nothing.
@@ -569,16 +743,6 @@ class OpalManager : public PObject
     );
 
 #if OPAL_VIDEO
-
-    /**Add video media formats available on a connection.
-
-       The default behaviour calls the OpalEndPoint function of the same name.
-      */
-    virtual void AddVideoMediaFormats(
-      OpalMediaFormatList & mediaFormats, ///<  Media formats to use
-      const OpalConnection * connection = NULL  ///<  Optional connection that is using formats
-    ) const;
-
     /**Create a PVideoInputDevice for a source media stream.
       */
     virtual PBoolean CreateVideoInputDevice(
@@ -622,13 +786,20 @@ class OpalManager : public PObject
 
     /**Call back for a media patch thread starting.
        This function is called within the context of the thread associated
-       with the media patch. It may be used to do any last checks on if the
-       patch should proceed.
+       with the media patch.
 
-       The default behaviour simply returns PTrue.
+       The default behaviour does nothing
       */
-    virtual PBoolean OnStartMediaPatch(
-      const OpalMediaPatch & patch     ///<  Media patch being started
+    virtual void OnStartMediaPatch(
+      OpalConnection & connection,  ///< Connection patch is in
+      OpalMediaPatch & patch        ///< Media patch being started
+    );
+
+    /**Call back when media stream patch thread stops.
+      */
+    virtual void OnStopMediaPatch(
+      OpalConnection & connection,  ///< Connection patch is in
+      OpalMediaPatch & patch        ///< Media Patch being stopped
     );
   //@}
 
@@ -713,7 +884,7 @@ class OpalManager : public PObject
        The "pattern" string regex is compared against routing strings that are built 
        as follows:
        
-                 a_party '\t' b_party
+                 a_party '\\t' b_party
 
        where:
 
@@ -725,8 +896,8 @@ class OpalManager : public PObject
 
        Note that all "pattern" strings have an implied '^' at the beginning and
        a '$' at the end. This forces the "pattern" to match the entire source string. 
-       For convenience, the sub-expression ".*\t" is inserted immediately after
-       any ':' character if no '\t' is present.
+       For convenience, the sub-expression ".*\\t" is inserted immediately after
+       any ':' character if no '\\t' is present.
 
        Route entries are stored and searched in the route table in the order they are added. 
        
@@ -756,17 +927,17 @@ class OpalManager : public PObject
        1) A local H.323 endpoint with with name of "myname@myhost.com" that receives a 
           call with a destination h323Id of "boris" would generate:
           
-                          "h323:myname@myhost.com\tboris"
+                          "h323:myname@myhost.com\\tboris"
 
        2) A local SIP endpoint with with name of "fred@nurk.com" that receives a 
           call with a destination of "sip:fred@nurk.com" would generate:
           
-                          "sip:fred@nurk.com\tsip:fred@nurk.com"
+                          "sip:fred@nurk.com\\tsip:fred@nurk.com"
 
        3) Using line 0 of a PhoneJACK handset with a serial # of 01AB3F4 to dial
           the digits 2, 6 and # would generate:
 
-                          "pots:Quicknet:01AB3F4:0\t26"
+                          "pots:Quicknet:01AB3F4:0\\t26"
 
 
        Destination meta-strings:
@@ -775,7 +946,7 @@ class OpalManager : public PObject
        The available meta-strings are:
        
          <da>    Replaced by the "b_party" string. For example
-                 "pc:.*\t.* = sip:<da>" directs calls to the SIP protocol. In
+                 "pc:.*\\t.* = sip:<da>" directs calls to the SIP protocol. In
                  this case there is a special condition where if the original
                  destination had a valid protocol, eg h323:fred.com, then
                  the entire string is replaced not just the <da> part.
@@ -833,7 +1004,7 @@ class OpalManager : public PObject
 
     /**Get the active route table for the manager.
       */
-    const RouteTable & GetRouteTable() const { return routeTable; }
+    const RouteTable & GetRouteTable() const { return m_routeTable; }
 
     /**Route the source address to a destination using the route table.
        The source parameter may be something like pots:vpb:1/2 or
@@ -849,6 +1020,40 @@ class OpalManager : public PObject
     );
   //@}
 
+#if OPAL_HAS_MIXER
+  /**@name Call recording */
+  //@{
+    /**Start recording a call.
+       Current version saves to a WAV file. It may either mix the receive and
+       transmit audio stream to a single mono file, or the streams are placed
+       into the left and right channels of a stereo WAV file.
+
+       Returns true if the call exists and there is no recording in progress
+               for the call.
+      */
+    virtual PBoolean StartRecording(
+      const PString & callToken,  ///< Call token for call to record
+      const PFilePath & filename, ///< File into which to record
+      const OpalRecordManager::Options & options = false ///< Record mixing options
+    );
+
+    /**Indicate if recording is currently active on call.
+      */
+    virtual bool IsRecording(
+      const PString & callToken   ///< Call token for call to check if recording
+    );
+
+    /** Stop a recording.
+        Returns true if the call does exists, that recording is active is
+                not indicated.
+      */
+    virtual bool StopRecording(
+      const PString & callToken   ///< Call token for call to stop recording
+    );
+
+  //@}
+#endif
+
   /**@name Member variable access */
   //@{
     /**Get the product info for all endpoints.
@@ -898,7 +1103,7 @@ class OpalManager : public PObject
 
     /**Set if should auto-start receive video channels on connection.
      */
-    void SetAutoStartReceiveVideo(bool can) { OpalMediaType::Video().GetDefinition()->SetAutoStart(CanAutoStartTransmitVideo() ? can ? OpalMediaType::ReceiveTransmit : OpalMediaType::Transmit : can ? OpalMediaType::Receive : OpalMediaType::OfferInactive); }
+    void SetAutoStartReceiveVideo(bool can) { OpalMediaType::Video().GetDefinition()->SetAutoStart(OpalMediaType::Receive, can); }
 
     /**See if should auto-start transmit video channels on connection.
      */
@@ -906,7 +1111,7 @@ class OpalManager : public PObject
 
     /**Set if should auto-start transmit video channels on connection.
      */
-    void SetAutoStartTransmitVideo(bool can) { OpalMediaType::Video().GetDefinition()->SetAutoStart(CanAutoStartReceiveVideo() ? can ? OpalMediaType::ReceiveTransmit : OpalMediaType::Receive : can ? OpalMediaType::Transmit : OpalMediaType::OfferInactive); }
+    void SetAutoStartTransmitVideo(bool can) { OpalMediaType::Video().GetDefinition()->SetAutoStart(OpalMediaType::Transmit, can); }
 
 #endif
 
@@ -1057,13 +1262,13 @@ class OpalManager : public PObject
      */
     void SetRtpIpTypeofService(unsigned tos) { rtpIpTypeofService = (BYTE)tos; }
 
-    /**Get the maximum RTP payload size.
+    /**Get the maximum transmitted RTP payload size.
        Defaults to maximum safe MTU size (576 bytes as per RFC879) minus the
        typical size of the IP, UDP an RTP headers.
       */
     PINDEX GetMaxRtpPayloadSize() const { return rtpPayloadSizeMax; }
 
-    /**Get the maximum RTP payload size.
+    /**Get the maximum transmitted RTP payload size.
        Defaults to maximum safe MTU size (576 bytes as per RFC879) minus the
        typical size of the IP, UDP an RTP headers.
       */
@@ -1072,6 +1277,18 @@ class OpalManager : public PObject
       bool mtu = false
     ) { rtpPayloadSizeMax = size - (mtu ? (20+16+12) : 0); }
 
+    /**Get the maximum received RTP packet size.
+       Defaults to 2048.
+      */
+    PINDEX GetMaxRtpPacketSize() const { return rtpPacketSizeMax; }
+
+    /**Get the maximum transmitted RTP payload size.
+       Defaults to 2048.
+      */
+    void SetMaxRtpPacketSize(
+      PINDEX size
+    ) { rtpPacketSizeMax = size; }
+
     /**Get the default maximum audio jitter delay parameter.
        Defaults to 50ms
      */
@@ -1107,10 +1324,16 @@ class OpalManager : public PObject
     );
 
     /**Get the default media format mask.
+       The is the default list of media format names to be removed from media
+       format lists bfeore use by a connection.
+       See OpalMediaFormatList::Remove() for more information.
      */
     const PStringArray & GetMediaFormatMask() const { return mediaFormatMask; }
 
     /**Set the default media format mask.
+       The is the default list of media format names to be removed from media
+       format lists bfeore use by a connection.
+       See OpalMediaFormatList::Remove() for more information.
      */
     void SetMediaFormatMask(
       const PStringArray & mask   //< New mask
@@ -1126,6 +1349,7 @@ class OpalManager : public PObject
      */
     const OpalSilenceDetector::Params & GetSilenceDetectParams() const { return silenceDetectParams; }
     
+#if OPAL_AEC
     /**Set the default parameters for the echo cancelation.
      */
     virtual void SetEchoCancelParams(
@@ -1135,6 +1359,7 @@ class OpalManager : public PObject
     /**Get the default parameters for the silence detector.
      */
     const OpalEchoCanceler::Params & GetEchoCancelParams() const { return echoCancelParams; }
+#endif
 
 #if OPAL_VIDEO
 
@@ -1221,55 +1446,10 @@ class OpalManager : public PObject
     // needs to be public for gcc 3.4
     void GarbageCollection();
 
-    /**Call back for a new connection has been constructed.
-       This is called after CreateConnection has returned a new connection.
-       It allows an application to make any custom adjustments to the
-       connection before it begins to process the protocol. behind it.
-      */
-    virtual void OnNewConnection(
-      OpalConnection & connection   ///< New connection just created
-    );
-
-    OpalRecordManager & GetRecordManager() const
-    { return *m_recordManager; }
-
-    /**Start recording a call.
-       Current version saves to a WAV file. It may either mix the receive and
-       transmit audio stream to a single mono file, or the streams are placed
-       into the left and right channels of a stereo WAV file.
-
-       Returns true if the call exists and there is no recording in progress
-               for the call.
-      */
-    virtual PBoolean StartRecording(
-      const PString & callToken,  ///< Call token for call to record
-      const PFilePath & filename, ///< File into which to record
-      bool mono = false           ///< Record as mono/stereo
-    );
-
-    /**Indicate if recording is currently active on call.
-      */
-    virtual bool IsRecording(
-      const PString & callToken   ///< Call token for call to check if recording
-    );
-
-    /** Stop a recording.
-        Returns true if the call does exists, that recording is active is
-                not indicated.
-      */
-    virtual bool StopRecording(
-      const PString & callToken   ///< Call token for call to stop recording
-    );
-
 #ifdef OPAL_ZRTP
     virtual bool GetZRTPEnabled() const;
 #endif
 
-    virtual void AddIMMediaFormats(
-      OpalMediaFormatList & mediaFormats,       ///<  Media formats to use
-      const OpalConnection * connection = NULL  ///<  Optional connection that is using formats
-    ) const;
-
     virtual void OnApplyStringOptions(
       OpalConnection & conn,
       OpalConnection::StringOptions & stringOptions
@@ -1284,6 +1464,7 @@ class OpalManager : public PObject
 
     BYTE          rtpIpTypeofService;
     PINDEX        rtpPayloadSizeMax;
+    PINDEX        rtpPacketSizeMax;
     unsigned      minAudioJitterDelay;
     unsigned      maxAudioJitterDelay;
     PStringArray  mediaFormatOrder;
@@ -1293,7 +1474,9 @@ class OpalManager : public PObject
     PString       ilsServer;
 
     OpalSilenceDetector::Params silenceDetectParams;
+#if OPAL_AEC
     OpalEchoCanceler::Params echoCancelParams;
+#endif
 
 #if OPAL_VIDEO
     PVideoDevice::OpenArgs videoInputDevice;
@@ -1341,8 +1524,8 @@ class OpalManager : public PObject
     PSTUNClient      * stun;
     InterfaceMonitor * interfaceMonitor;
 
-    RouteTable routeTable;
-    PMutex     routeTableMutex;
+    RouteTable m_routeTable;
+    PMutex     m_routeMutex;
 
     // Dynamic variables
     PReadWriteMutex     endpointsMutex;
@@ -1359,6 +1542,8 @@ class OpalManager : public PObject
         OpalManager & manager;
     } activeCalls;
 
+    PSafeDictionary<PString, OpalPresentity> m_presentities;
+
     PAtomicInteger m_clearingAllCallsCount;
     PMutex         m_clearingAllCallsMutex;
     PSyncPoint     m_allCallsCleared;
@@ -1372,10 +1557,15 @@ class OpalManager : public PObject
     bool zrtpEnabled;
 #endif
 
-    OpalRecordManager * m_recordManager;
-
     friend OpalCall::OpalCall(OpalManager & mgr);
-    friend void OpalCall::OnReleased(OpalConnection & connection);
+    friend void OpalCall::InternalOnClear();
+
+  private:
+    P_REMOVE_VIRTUAL(OpalCall *,CreateCall(), 0);
+    P_REMOVE_VIRTUAL(PBoolean, OnIncomingConnection(OpalConnection &, unsigned), false);
+    P_REMOVE_VIRTUAL(PBoolean, OnIncomingConnection(OpalConnection &), false);
+    P_REMOVE_VIRTUAL(PBoolean, OnStartMediaPatch(const OpalMediaPatch &), false);
+    P_REMOVE_VIRTUAL_VOID(AdjustMediaFormats(const OpalConnection &, OpalMediaFormatList &) const);
 };
 
 
