@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22902 $
+ * $Revision: 24154 $
  * $Author: rjongbloed $
- * $Date: 2009-06-15 23:58:09 -0500 (Mon, 15 Jun 2009) $
+ * $Date: 2010-03-23 20:56:23 -0500 (Tue, 23 Mar 2010) $
  */
 
 #ifndef OPAL_SIP_SDP_H
@@ -80,7 +80,7 @@ class SDPMediaFormat : public PObject
 
     RTP_DataFrame::PayloadTypes GetPayloadType() const { return payloadType; }
 
-    PString GetEncodingName() const         { return encodingName; }
+    const PCaselessString & GetEncodingName() const { return encodingName; }
     void SetEncodingName(const PString & v) { encodingName = v; }
 
     void SetFMTP(const PString & _fmtp); 
@@ -91,24 +91,24 @@ class SDPMediaFormat : public PObject
 
     void SetParameters(const PString & v) { parameters = v; }
 
-    void SetPacketTime(const PString & optionName, unsigned ptime);
-
     const OpalMediaFormat & GetMediaFormat() const;
+    OpalMediaFormat & GetWritableMediaFormat();
 
     bool PreEncode();
     bool PostDecode(unsigned bandwidth);
 
   protected:
-    void InitialiseMediaFormat();
+    void InitialiseMediaFormat(OpalMediaFormat & mediaFormat) const;
+    void SetMediaFormatOptions(OpalMediaFormat & mediaFormat) const;
 
-    OpalMediaFormat mediaFormat;
+    OpalMediaFormat m_mediaFormat;
 
     SDPMediaDescription & m_parent;
     RTP_DataFrame::PayloadTypes payloadType;
     unsigned clockRate;
-    PString encodingName;
+    PCaselessString encodingName;
     PString parameters;
-    PString fmtp;
+    PString m_fmtp;
 };
 
 PLIST(SDPMediaFormatList, SDPMediaFormat);
@@ -130,7 +130,8 @@ class SDPMediaDescription : public PObject
     };
 
     SDPMediaDescription(
-      const OpalTransportAddress & address
+      const OpalTransportAddress & address,
+      const OpalMediaType & mediaType
     );
 
     virtual bool PreEncode();
@@ -174,8 +175,6 @@ class SDPMediaDescription : public PObject
 
     virtual const SDPBandwidth & GetBandwidth() const { return bandwidth; }
 
-    virtual void RemoveSDPMediaFormat(const SDPMediaFormat & sdpMediaFormat);
-
     virtual void CreateSDPMediaFormats(const PStringArray & tokens);
     virtual SDPMediaFormat * CreateSDPMediaFormat(const PString & portString) = 0;
 
@@ -183,9 +182,11 @@ class SDPMediaDescription : public PObject
 
     virtual void ProcessMediaOptions(SDPMediaFormat & sdpFormat, const OpalMediaFormat & mediaFormat);
 
+    unsigned GetPTime () const { return ptime; }
+    unsigned GetMaxPTime () const { return maxptime; }
+
   protected:
     virtual SDPMediaFormat * FindFormat(PString & str) const;
-    virtual void SetPacketTime(const PString & optionName, const PString & value);
 
     OpalTransportAddress transportAddress;
     Direction direction;
@@ -195,6 +196,8 @@ class SDPMediaDescription : public PObject
 
     SDPMediaFormatList formats;
     SDPBandwidth       bandwidth;
+    unsigned           ptime;
+    unsigned           maxptime;
 };
 
 PARRAY(SDPMediaDescriptionArray, SDPMediaDescription);
@@ -208,7 +211,7 @@ class SDPRTPAVPMediaDescription : public SDPMediaDescription
 {
   PCLASSINFO(SDPRTPAVPMediaDescription, SDPMediaDescription);
   public:
-    SDPRTPAVPMediaDescription(const OpalTransportAddress & address);
+    SDPRTPAVPMediaDescription(const OpalTransportAddress & address, const OpalMediaType & mediaType);
     virtual PCaselessString GetSDPTransportType() const;
     virtual SDPMediaFormat * CreateSDPMediaFormat(const PString & portString);
     virtual PString GetSDPPortList() const;
@@ -243,6 +246,8 @@ class SDPVideoMediaDescription : public SDPRTPAVPMediaDescription
     SDPVideoMediaDescription(const OpalTransportAddress & address);
     virtual PString GetSDPMediaType() const;
     virtual bool PreEncode();
+    virtual bool PrintOn(ostream & str, const PString & connectString) const;
+    void SetAttribute(const PString & attr, const PString & value);
 };
 
 /////////////////////////////////////////////////////////
@@ -310,6 +315,8 @@ class SDPSessionDescription : public PObject
     unsigned GetBandwidth(const PString & type) const { return bandwidth[type]; }
     void SetBandwidth(const PString & type, unsigned value) { bandwidth[type] = value; }
 
+    OpalMediaFormatList GetMediaFormats() const;
+
     static const PString & ConferenceTotalBandwidthType();
     static const PString & ApplicationSpecificBandwidthType();
     static const PString & TransportIndependentBandwidthType(); // RFC3890
