@@ -22,9 +22,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22220 $
+ * $Revision: 23863 $
  * $Author: rjongbloed $
- * $Date: 2009-03-18 17:51:48 -0500 (Wed, 18 Mar 2009) $
+ * $Date: 2009-12-09 20:42:20 -0600 (Wed, 09 Dec 2009) $
  */
 
 #include <ptlib.h>
@@ -48,8 +48,9 @@ static PString GetCapability(const std::vector<bool> & capabilitySet)
 {
   PStringStream str;
 
+  PINDEX last = capabilitySet.size()-1;
   PINDEX i = 0;
-  while (i < 255) {
+  while (i < last) {
     if (!capabilitySet[i])
       i++;
     else {
@@ -68,29 +69,38 @@ static PString GetCapability(const std::vector<bool> & capabilitySet)
 }
 
 
-static void SetCapability(const PString & codes, std::vector<bool> & capabilitySet)
+static void SetCapability(const PString & codes, std::vector<bool> & xxCapabilitySet, bool merge)
 {
+  std::vector<bool> capabilitySet;
+  capabilitySet.resize(xxCapabilitySet.size());
+ 
   if (codes.IsEmpty()) {
-    SetCapability("0-15", capabilitySet); // RFC specified default
-    return;
+    // RFC specified default: 0-15
+    for (size_t i = 0; i < 15; ++i)
+      capabilitySet[i] = true;
   }
-
-  capabilitySet.resize(0);
-  capabilitySet.resize(256);
-
-  PStringArray tokens = codes.Tokenise(',');
-  for (PINDEX i = 0; i < tokens.GetSize(); ++i) {
-    PString token = tokens[i];
-    unsigned code = token.AsUnsigned();
-    if (code < 256) {
-      PINDEX dash = token.Find('-');
-      unsigned end = dash == P_MAX_INDEX ? code : token.Mid(dash+1).AsUnsigned();
-      if (end > 255)
-        end = 255;
-      while (code <= end)
-        capabilitySet[code++] = true;
+  else if (codes != "-") {       // Allow for an empty set
+    PStringArray tokens = codes.Tokenise(',');
+    for (PINDEX i = 0; i < tokens.GetSize(); ++i) {
+      PString token = tokens[i];
+      unsigned code = token.AsUnsigned();
+      if (code < capabilitySet.size()) {
+        PINDEX dash = token.Find('-');
+        unsigned end = dash == P_MAX_INDEX ? code : token.Mid(dash+1).AsUnsigned();
+        if (end >= capabilitySet.size())
+          end = capabilitySet.size()-1;
+        while (code <= end)
+          capabilitySet[code++] = true;
+      }
     }
   }
+
+  if (merge) {
+    for (size_t i = 0; i < capabilitySet.size(); ++i)
+      xxCapabilitySet[i] = xxCapabilitySet[i] && capabilitySet[i];
+  }
+  else
+    xxCapabilitySet = capabilitySet;
 }
 
 
@@ -133,6 +143,7 @@ OpalRFC2833Proto::OpalRFC2833Proto(OpalRTPConnection & conn, const PNotifier & r
   m_asyncTransmitTimer.SetNotifier(PCREATE_NOTIFIER(AsyncTimeout));
   m_asyncDurationTimer.SetNotifier(PCREATE_NOTIFIER(AsyncTimeout));
 
+  m_rxCapabilitySet.resize(256);
   SetRxCapability(fmt.GetOptionString("FMTP", "0-15"));
   m_txCapabilitySet = m_rxCapabilitySet;
 }
@@ -203,9 +214,16 @@ void OpalRFC2833Proto::SendAsyncFrame()
   // be thread safe
   PWaitAndSignal mutex(m_mutex);
 
-  // if transmittter is ever in this state, then stop the duration timer
-  if (m_payloadType == RTP_DataFrame::IllegalPayloadType)
+  if (m_rtpSession == NULL) {
+    PTRACE(2, "RFC2833\tCannot send as not RTP session attached");
     m_transmitState = TransmitIdle;
+  }
+
+  // if transmittter is ever in this state, then stop the duration timer
+  if (m_payloadType == RTP_DataFrame::IllegalPayloadType) {
+    PTRACE(2, "RFC2833\tNo payload type for sent packet.");
+    m_transmitState = TransmitIdle;
+  }
 
   if (m_transmitState == TransmitIdle) {
     m_asyncDurationTimer.Stop(false);
@@ -265,18 +283,24 @@ void OpalRFC2833Proto::SendAsyncFrame()
   payload[2] = (BYTE)(m_transmitDuration >> 8);
   payload[3] = (BYTE) m_transmitDuration;
 
-  if (m_rtpSession != NULL) {
-    if (!m_rewriteTransmitTimestamp)
-      frame.SetTimestamp(m_transmitTimestamp);
-    m_rtpSession->WriteOOBData(frame, m_rewriteTransmitTimestamp);
-    if (m_rewriteTransmitTimestamp) {
-      m_transmitTimestamp        = frame.GetTimestamp();
-      m_rewriteTransmitTimestamp = false;
-    } 
+  if (!m_rewriteTransmitTimestamp)
+    frame.SetTimestamp(m_transmitTimestamp);
+
+  if (!m_rtpSession->WriteOOBData(frame, m_rewriteTransmitTimestamp)) {
+    PTRACE(3, "RFC2833\tTransmission stopped by RTP session");
+    // Abort further transmission
+    m_transmitState = TransmitIdle;
+    m_asyncDurationTimer.Stop(false);
   }
 
-  PTRACE(4, "RFC2833\tSending " << ((payload[1] & 0x80) ? "end" : "tone") << ": code=" << (unsigned)m_transmitCode
-         << ", dur=" << m_transmitDuration << ", ts=" << frame.GetTimestamp() << ", mkr=" << frame.GetMarker());
+  if (m_rewriteTransmitTimestamp) {
+    m_transmitTimestamp        = frame.GetTimestamp();
+    m_rewriteTransmitTimestamp = false;
+  } 
+
+  PTRACE(frame.GetMarker() ? 3 : 4,
+         "RFC2833\tSent " << ((payload[1] & 0x80) ? "end" : "tone") << ": code=" << (unsigned)m_transmitCode <<
+         ", dur=" << m_transmitDuration << ", ts=" << frame.GetTimestamp() << ", mkr=" << frame.GetMarker());
 }
 
 
@@ -292,15 +316,17 @@ PString OpalRFC2833Proto::GetRxCapability() const
 }
 
 
-void OpalRFC2833Proto::SetTxCapability(const PString & codes)
+void OpalRFC2833Proto::SetTxCapability(const PString & codes, bool merge)
 {
-  SetCapability(codes, m_txCapabilitySet);
+  PTRACE(4, "RFC2833\tTx capability " << (merge ? "merged with" : "set to") << " \"" << codes << '"');
+  SetCapability(codes, m_txCapabilitySet, merge);
 }
 
 
 void OpalRFC2833Proto::SetRxCapability(const PString & codes)
 {
-  SetCapability(codes, m_rxCapabilitySet);
+  PTRACE(4, "RFC2833\tRx capability set to \"" << codes << '"');
+  SetCapability(codes, m_rxCapabilitySet, false);
 }
 
 
