@@ -26,9 +26,9 @@
  *
  * Contributor(s): Many thanks to Simon Horne.
  *
- * $Revision: 22361 $
+ * $Revision: 23224 $
  * $Author: rjongbloed $
- * $Date: 2009-04-01 07:06:50 +0000 (Wed, 01 Apr 2009) $
+ * $Date: 2009-08-19 23:26:44 -0500 (Wed, 19 Aug 2009) $
  */
 
 #include <ptlib.h>
@@ -57,65 +57,61 @@
 /////////////////////////////////////////////////////////////////////////////
 
 H323EndPoint::H323EndPoint(OpalManager & manager)
-  : OpalRTPEndPoint(manager, "h323", CanTerminateCall),
-    m_bH245Disabled(PFalse),
-    signallingChannelCallTimeout(0, 0, 1),  // Minutes
-    controlChannelStartTimeout(0, 0, 2),    // Minutes
-    endSessionTimeout(0, 10),               // Seconds
-    masterSlaveDeterminationTimeout(0, 30), // Seconds
-    capabilityExchangeTimeout(0, 30),       // Seconds
-    logicalChannelTimeout(0, 30),           // Seconds
-    requestModeTimeout(0, 30),              // Seconds
-    roundTripDelayTimeout(0, 10),           // Seconds
-    roundTripDelayRate(0, 0, 1),            // Minutes
-    gatekeeperRequestTimeout(0, 5),         // Seconds
-    rasRequestTimeout(0, 3),                // Seconds
-    callTransferT1(0,10),                   // Seconds
-    callTransferT2(0,10),                   // Seconds
-    callTransferT3(0,10),                   // Seconds
-    callTransferT4(0,10),                   // Seconds
-    callIntrusionT1(0,30),                  // Seconds
-    callIntrusionT2(0,30),                  // Seconds
-    callIntrusionT3(0,30),                  // Seconds
-    callIntrusionT4(0,30),                  // Seconds
-    callIntrusionT5(0,10),                  // Seconds
-    callIntrusionT6(0,10)                   // Seconds
+  : OpalRTPEndPoint(manager, "h323", CanTerminateCall)
+  , autoCallForward(true)
+  , disableFastStart(false)
+  , disableH245Tunneling(false)
+  , disableH245inSetup(false)
+  , m_bH245Disabled(false)
+  , canDisplayAmountString(false)
+  , canEnforceDurationLimit(true)
 #if OPAL_H450
-    ,nextH450CallIdentity(0)
+  , callIntrusionProtectionLevel(3) //H45011_CIProtectionLevel::e_fullProtection;
+#endif
+  , terminalType(e_TerminalOnly)
+#if OPAL_H239
+  , m_defaultH239Control(false)
+#endif
+  , clearCallOnRoundTripFail(false)
+  , signallingChannelCallTimeout(0, 0, 1)  // Minutes
+  , controlChannelStartTimeout(0, 0, 2)    // Minutes
+  , endSessionTimeout(0, 10)               // Seconds
+  , masterSlaveDeterminationTimeout(0, 30) // Seconds
+  , masterSlaveDeterminationRetries(10)
+  , capabilityExchangeTimeout(0, 30)       // Seconds
+  , logicalChannelTimeout(0, 30)           // Seconds
+  , requestModeTimeout(0, 30)              // Seconds
+  , roundTripDelayTimeout(0, 10)           // Seconds
+  , roundTripDelayRate(0, 0, 1)            // Minutes
+  , gatekeeperRequestTimeout(0, 5)         // Seconds
+  , gatekeeperRequestRetries(2)
+  , rasRequestTimeout(0, 3)                // Seconds
+  , rasRequestRetries(2)
+  , registrationTimeToLive(0, 0, 10)       // Minutes
+  , sendGRQ(true)
+  , callTransferT1(0,10)                   // Seconds
+  , callTransferT2(0,10)                   // Seconds
+  , callTransferT3(0,10)                   // Seconds
+  , callTransferT4(0,10)                   // Seconds
+  , callIntrusionT1(0,30)                  // Seconds
+  , callIntrusionT2(0,30)                  // Seconds
+  , callIntrusionT3(0,30)                  // Seconds
+  , callIntrusionT4(0,30)                  // Seconds
+  , callIntrusionT5(0,10)                  // Seconds
+  , callIntrusionT6(0,10)                   // Seconds
+  , gatekeeper(NULL)
+#if OPAL_H450
+  , nextH450CallIdentity(0)
+#endif
+#if OPAL_H460
+  , disableH460(false)
 #endif
 {
-  // Set port in OpalEndPoint class
-  defaultSignalPort = 1720;
+  defaultSignalPort = 1720; // Set port in OpalEndPoint class
 
   localAliasNames.AppendString(defaultLocalPartyName);
 
-  m_bH245Disabled = PFalse;
-  autoCallForward = PTrue;
-  disableFastStart = PFalse;
-  disableH245Tunneling = PFalse;
-  disableH245inSetup = PFalse;
-  canDisplayAmountString = PFalse;
-  canEnforceDurationLimit = PTrue;
-
-#if OPAL_H450
-  callIntrusionProtectionLevel = 3; //H45011_CIProtectionLevel::e_fullProtection;
-#endif
-
-  terminalType = e_TerminalOnly;
-  clearCallOnRoundTripFail = PFalse;
-
-  masterSlaveDeterminationRetries = 10;
-  gatekeeperRequestRetries = 2;
-  rasRequestRetries = 2;
-  sendGRQ = PTrue;
-
-  gatekeeper = NULL;
-
   secondaryConnectionsActive.DisallowDeleteObjects();
-  
-#if OPAL_H460
-  disableH460 = FALSE;
-#endif
 
   manager.AttachEndPoint(this, "h323s");
 
@@ -130,10 +126,12 @@ H323EndPoint::~H323EndPoint()
 
 void H323EndPoint::ShutDown()
 {
-  OpalEndPoint::ShutDown();
-
-  // And shut down the gatekeeper (if there was one)
+  /* Unregister request needs/depends OpalEndpoint listeners object, so shut
+     down the gatekeeper (if there was one) before cleaning up the OpalEndpoint
+     object which kills the listeners. */
   RemoveGatekeeper();
+
+  OpalEndPoint::ShutDown();
 }
 
 
@@ -156,25 +154,32 @@ void H323EndPoint::SetEndpointTypeInfo(H225_EndpointType & info) const
     case e_TerminalAndMC :
       info.IncludeOptionalField(H225_EndpointType::e_terminal);
       break;
+
     case e_GatewayOnly :
     case e_GatewayAndMC :
     case e_GatewayAndMCWithDataMP :
     case e_GatewayAndMCWithAudioMP :
     case e_GatewayAndMCWithAVMP :
       info.IncludeOptionalField(H225_EndpointType::e_gateway);
+      if (SetGatewaySupportedProtocol(info.m_gateway.m_protocol))
+        info.m_gateway.IncludeOptionalField(H225_GatewayInfo::e_protocol);
       break;
+
     case e_GatekeeperOnly :
     case e_GatekeeperWithDataMP :
     case e_GatekeeperWithAudioMP :
     case e_GatekeeperWithAVMP :
       info.IncludeOptionalField(H225_EndpointType::e_gatekeeper);
       break;
+
     case e_MCUOnly :
     case e_MCUWithDataMP :
     case e_MCUWithAudioMP :
     case e_MCUWithAVMP :
       info.IncludeOptionalField(H225_EndpointType::e_mcu);
-      info.m_mc = PTrue;
+      info.m_mc = true;
+      if (SetGatewaySupportedProtocol(info.m_mcu.m_protocol))
+        info.m_mcu.IncludeOptionalField(H225_McuInfo::e_protocol);
   }
 }
 
@@ -201,6 +206,36 @@ void H323EndPoint::SetH221NonStandardInfo(H225_H221NonStandard & info) const
 }
 
 
+bool H323EndPoint::SetGatewaySupportedProtocol(H225_ArrayOf_SupportedProtocols & protocols) const
+{
+  PStringList prefixes;
+
+  if (!OnSetGatewayPrefixes(prefixes))
+    return false;
+
+  PINDEX count = protocols.GetSize();
+  protocols.SetSize(count+1);
+
+  protocols[count].SetTag(H225_SupportedProtocols::e_h323);
+  H225_H323Caps & caps = protocols[count];
+
+  caps.IncludeOptionalField(H225_H323Caps::e_supportedPrefixes);
+  H225_ArrayOf_SupportedPrefix & supportedPrefixes = caps.m_supportedPrefixes;	
+  supportedPrefixes.SetSize(prefixes.GetSize());
+
+  for (PINDEX i = 0; i < prefixes.GetSize(); i++)
+    H323SetAliasAddress(prefixes[i], supportedPrefixes[i].m_prefix);
+
+  return true;
+}
+
+
+bool H323EndPoint::OnSetGatewayPrefixes(PStringList & prefixes) const
+{
+	return false;
+}
+
+
 H323Capability * H323EndPoint::FindCapability(const H245_Capability & cap) const
 {
   return GetCapabilities().FindCapability(cap);
@@ -440,17 +475,14 @@ H235Authenticators H323EndPoint::CreateAuthenticators()
 }
 
 
-PBoolean H323EndPoint::MakeConnection(OpalCall & call,
-                                  const PString & remoteParty,
-                                  void * userData,
-                            unsigned int options,
-                            OpalConnection::StringOptions * stringOptions)
+PSafePtr<OpalConnection> H323EndPoint::MakeConnection(OpalCall & call,
+                                                 const PString & remoteParty,
+                                                          void * userData,
+                                                    unsigned int options,
+                                 OpalConnection::StringOptions * stringOptions)
 {
-  if ((remoteParty.NumCompare("h323:") != EqualTo) && (remoteParty.NumCompare("h323s") != EqualTo))
-    return false;
-
   if (listeners.IsEmpty())
-    return false;
+    return NULL;
 
   PTRACE(3, "H323\tMaking call to: " << remoteParty);
   return InternalMakeCall(call,
@@ -565,12 +597,7 @@ PBoolean H323EndPoint::SetupTransfer(const PString & oldToken,
   call.CloseMediaStreams();
 
   PTRACE(3, "H323\tTransferring call to: " << remoteParty);
-  PBoolean ok = InternalMakeCall(call,
-			     oldToken,
-			     callIdentity,
-			     UINT_MAX,
-			     remoteParty,
-			     userData);
+  PBoolean ok = InternalMakeCall(call, oldToken, callIdentity, UINT_MAX, remoteParty, userData) != NULL;
   call.OnReleased(*otherConnection);
   otherConnection->Release(OpalConnection::EndedByCallForwarded);
 
@@ -578,25 +605,29 @@ PBoolean H323EndPoint::SetupTransfer(const PString & oldToken,
 }
 
 
-PBoolean H323EndPoint::InternalMakeCall(OpalCall & call,
-                                    const PString & existingToken,
+H323Connection * H323EndPoint::InternalMakeCall(OpalCall & call,
+                                           const PString & existingToken,
 #if OPAL_H450
-                                    const PString & callIdentity,
-                                           unsigned capabilityLevel,
+                                           const PString & callIdentity,
+                                                  unsigned capabilityLevel,
 #else
-                                    const PString & ,
-                                           unsigned ,
+                                           const PString & ,
+                                                  unsigned ,
 #endif
-                                    const PString & remoteParty,
-                                             void * userData,
-                                       unsigned int options,
-                                OpalConnection::StringOptions * stringOptions)
+                                           const PString & remoteParty,
+                                                    void * userData,
+                                              unsigned int options,
+                           OpalConnection::StringOptions * stringOptions)
 {
+  OpalConnection::StringOptions localStringOptions;
+  if (stringOptions == NULL)
+    stringOptions = &localStringOptions;
+
   PString alias;
   H323TransportAddress address;
-  if (!ParsePartyName(remoteParty, alias, address)) {
+  if (!ParsePartyName(remoteParty, alias, address, stringOptions)) {
     PTRACE(2, "H323\tCould not parse \"" << remoteParty << '"');
-    return PFalse;
+    return NULL;
   }
 
   // Restriction: the call must be made on the same local interface as the one
@@ -605,7 +636,7 @@ PBoolean H323EndPoint::InternalMakeCall(OpalCall & call,
   if (gatekeeper != NULL)
     transport = gatekeeper->GetTransport().GetLocalAddress().CreateTransport(
                                           *this, OpalTransportAddress::Streamed);
-  else if (stringOptions == NULL || !stringOptions->Contains(OPAL_OPT_INTERFACE))
+  else if (!stringOptions->Contains(OPAL_OPT_INTERFACE))
     transport = address.CreateTransport(*this, OpalTransportAddress::NoBinding);
   else {
     OpalTransportAddress localInterface = (*stringOptions)[OPAL_OPT_INTERFACE];
@@ -614,7 +645,7 @@ PBoolean H323EndPoint::InternalMakeCall(OpalCall & call,
 
   if (transport == NULL) {
     PTRACE(1, "H323\tInvalid transport in \"" << remoteParty << '"');
-    return PFalse;
+    return NULL;
   }
 
   inUseFlag.Wait();
@@ -629,7 +660,7 @@ PBoolean H323EndPoint::InternalMakeCall(OpalCall & call,
   H323Connection * connection = CreateConnection(call, newToken, userData, *transport, alias, address, NULL, options, stringOptions);
   if (!AddConnection(connection)) {
     PTRACE(1, "H225\tEndpoint could not create connection, aborting setup.");
-    return PFalse;
+    return NULL;
   }
 
   inUseFlag.Signal();
@@ -653,7 +684,7 @@ PBoolean H323EndPoint::InternalMakeCall(OpalCall & call,
   if (call.GetConnection(0) == (OpalConnection*)connection || !existingToken.IsEmpty())
     connection->SetUpConnection();
 
-  return PTrue;
+  return connection;
 }
 
 
@@ -695,7 +726,7 @@ PBoolean H323EndPoint::IntrudeCall(const PString & remoteParty,
   if (call == NULL)
     return false;
 
-  return InternalMakeCall(*call, PString::Empty(), PString::Empty(), capabilityLevel, remoteParty, userData);
+  return InternalMakeCall(*call, PString::Empty(), PString::Empty(), capabilityLevel, remoteParty, userData) != NULL;
 }
 
 #endif
@@ -724,11 +755,15 @@ void H323EndPoint::OnReceivedInitiateReturnError()
 }
 
 PBoolean H323EndPoint::ParsePartyName(const PString & remoteParty,
-                                  PString & alias,
-                                  H323TransportAddress & address)
+                                            PString & alias,
+                               H323TransportAddress & address,
+                       OpalConnection::StringOptions * stringOptions)
 {
   PURL url(remoteParty, GetPrefixName()); // Parses as per RFC3508
 
+  if (stringOptions != NULL)
+    stringOptions->ExtractFromURL(url);
+
 #if OPAL_PTLIB_DNS
 
   // if there is no gatekeeper, try altarnate address lookup methods
@@ -1017,12 +1052,12 @@ PBoolean H323EndPoint::ForwardConnection(H323Connection & connection,
                                      const PString & forwardParty,
                                      const H323SignalPDU & /*pdu*/)
 {
-  if (!InternalMakeCall(connection.GetCall(),
-                        connection.GetCallToken(),
-                        PString::Empty(),
-                        UINT_MAX,
-                        forwardParty,
-                        NULL))
+  if (InternalMakeCall(connection.GetCall(),
+                       connection.GetCallToken(),
+                       PString::Empty(),
+                       UINT_MAX,
+                       forwardParty,
+                       NULL) == NULL)
     return PFalse;
 
   connection.SetCallEndReason(H323Connection::EndedByCallForwarded);
