@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23153 $
+ * $Revision: 24261 $
  * $Author: rjongbloed $
- * $Date: 2009-07-29 02:22:00 +0000 (Wed, 29 Jul 2009) $
+ * $Date: 2010-04-22 23:24:13 -0500 (Thu, 22 Apr 2010) $
  */
 
 #include <ptlib.h>
@@ -47,19 +47,6 @@
 
 #define new PNEW
 
-namespace PWLibStupidLinkerHacks {
-extern int opalLoader;
-
-static class InstantiateMe
-{
-  public:
-    InstantiateMe()
-    {
-      opalLoader = 1;
-    }
-} instance;
-
-}; // namespace PWLibStupidLinkerHacks
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -70,17 +57,17 @@ static class InstantiateMe
                                     encodingName, frameSize, frameTime, rxFrames, txFrames, maxFrames, clock); \
     return name; \
   }
-
-AUDIO_FORMAT(PCM16,          MaxPayloadType, "",     16, 8,  240,  0, 256,  8000);
-AUDIO_FORMAT(PCM16_16KHZ,    MaxPayloadType, "",     32,16,  240,  0, 256, 16000);
-AUDIO_FORMAT(PCM16_32KHZ,    MaxPayloadType, "",     64,32,  240,  0, 256, 32000);
-AUDIO_FORMAT(PCM16_48KHZ,    MaxPayloadType, "",     96,48,  240,  0, 256, 48000);
-AUDIO_FORMAT(L16_MONO_8KHZ,  L16_Mono,       "L16",  16, 8,  240, 30, 256,  8000);
-AUDIO_FORMAT(L16_MONO_16KHZ, L16_Mono,       "L16",  32,16,  240, 30, 256, 16000);
-AUDIO_FORMAT(L16_MONO_32KHZ, L16_Mono,       "L16",  64,32,  240, 30, 256, 32000);
-AUDIO_FORMAT(L16_MONO_48KHZ, L16_Mono,       "L16",  96,48,  240, 30, 256, 48000);
-AUDIO_FORMAT(G711_ULAW_64K,  PCMU,           "PCMU",  8, 8,  240, 20, 256,  8000);
-AUDIO_FORMAT(G711_ALAW_64K,  PCMA,           "PCMA",  8, 8,  240, 20, 256,  8000);
+//           name            rtpPayloadType  encodingName frameSize frameTime rxFrames txFrames maxFrames clock
+AUDIO_FORMAT(PCM16,          MaxPayloadType, "",          16,        8,       240,      0,      256,       8000);
+AUDIO_FORMAT(PCM16_16KHZ,    MaxPayloadType, "",          32,       16,       240,      0,      256,      16000);
+AUDIO_FORMAT(PCM16_32KHZ,    MaxPayloadType, "",          64,       32,       240,      0,      256,      32000);
+AUDIO_FORMAT(PCM16_48KHZ,    MaxPayloadType, "",          96,       48,       240,      0,      256,      48000);
+AUDIO_FORMAT(L16_MONO_8KHZ,  L16_Mono,       "L16",       16,        8,       240,     30,      256,       8000);
+AUDIO_FORMAT(L16_MONO_16KHZ, L16_Mono,       "L16",       32,       16,       240,     30,      256,      16000);
+AUDIO_FORMAT(L16_MONO_32KHZ, L16_Mono,       "L16",       64,       32,       240,     30,      256,      32000);
+AUDIO_FORMAT(L16_MONO_48KHZ, L16_Mono,       "L16",       96,       48,       240,     30,      256,      48000);
+AUDIO_FORMAT(G711_ULAW_64K,  PCMU,           "PCMU",       8,        8,       240,     20,      256,       8000);
+AUDIO_FORMAT(G711_ALAW_64K,  PCMA,           "PCMA",       8,        8,       240,     20,      256,       8000);
 
 
 class OpalStereoAudioFormat : public OpalAudioFormat
@@ -128,7 +115,7 @@ const OpalMediaFormat & GetOpalRFC2833()
                         true,   // Needs jitter
                         32*(1000/50), // bits/sec  (32 bits every 50ms)
                         4,      // bytes/frame
-                        150*8,  // 150 millisecond
+                        10*8,   // 10 millisecond
                         OpalMediaFormat::AudioClockRate)
     {
       AddOption(new OpalMediaOptionString("FMTP", false, "0-16,32,36"));
@@ -148,7 +135,7 @@ const OpalMediaFormat & GetOpalCiscoNSE()
                         true,   // Needs jitter
                         32*(1000/50), // bits/sec  (32 bits every 50ms)
                         4,      // bytes/frame
-                        150*8,  // 150 millisecond
+                        10*8,   // 10 millisecond
                         OpalMediaFormat::AudioClockRate)
     {
       AddOption(new OpalMediaOptionString("FMTP", false, "192,193"));
@@ -241,13 +228,15 @@ bool OpalMediaOption::Merge(const OpalMediaOption & option)
     case EqualMerge :
       if (CompareValue(option) == EqualTo)
         return true;
-      PTRACE(2, "MediaFormat\tMerge of media option \"" << m_name << "\" failed, required to be equal.");
+      PTRACE(2, "MediaFormat\tMerge of media option \"" << m_name << "\" failed, "
+                "required to be equal: \"" << *this << "\"!=\"" << option << '"');
       return false;
 
     case NotEqualMerge :
       if (CompareValue(option) != EqualTo)
         return true;
-      PTRACE(2, "MediaFormat\tMerge of media option \"" << m_name << "\" failed, required to be not equal.");
+      PTRACE(2, "MediaFormat\tMerge of media option \"" << m_name << "\" failed, "
+                "required to be not equal: \"" << *this << "\"==\"" << option << '"');
       return false;
 
     case AlwaysMerge :
@@ -268,6 +257,28 @@ bool OpalMediaOption::Merge(const OpalMediaOption & option)
 }
 
 
+bool OpalMediaOption::ValidateMerge(const OpalMediaOption & option) const
+{
+  switch (m_merge) {
+    case EqualMerge :
+      if (CompareValue(option) == EqualTo)
+        return true;
+      break;
+
+    case NotEqualMerge :
+      if (CompareValue(option) != EqualTo)
+        return true;
+      break;
+
+    default :
+      return true;
+  }
+
+  PTRACE(2, "MediaFormat\tValidation of merge for media option \"" << m_name << "\" failed.");
+  return false;
+}
+
+
 PString OpalMediaOption::AsString() const
 {
   PStringStream strm;
@@ -710,7 +721,7 @@ void OpalMediaFormat::Construct(OpalMediaFormatInternal * info)
 
 OpalMediaFormat & OpalMediaFormat::operator=(RTP_DataFrame::PayloadTypes pt)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
 
   PWaitAndSignal mutex(GetMediaFormatsListMutex());
   const OpalMediaFormatList & registeredFormats = GetMediaFormatsList();
@@ -727,14 +738,14 @@ OpalMediaFormat & OpalMediaFormat::operator=(RTP_DataFrame::PayloadTypes pt)
 
 OpalMediaFormat & OpalMediaFormat::operator=(const char * wildcard)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   return operator=(PString(wildcard));
 }
 
 
 OpalMediaFormat & OpalMediaFormat::operator=(const PString & wildcard)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   PWaitAndSignal mutex(GetMediaFormatsListMutex());
   const OpalMediaFormatList & registeredFormats = GetMediaFormatsList();
 
@@ -750,7 +761,7 @@ OpalMediaFormat & OpalMediaFormat::operator=(const PString & wildcard)
 
 void OpalMediaFormat::CloneContents(const OpalMediaFormat * c)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
 
   m_info = (OpalMediaFormatInternal *)c->m_info->Clone();
   m_info->options.MakeUnique();
@@ -759,28 +770,28 @@ void OpalMediaFormat::CloneContents(const OpalMediaFormat * c)
 
 void OpalMediaFormat::CopyContents(const OpalMediaFormat & c)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   m_info = c.m_info;
 }
 
 
 void OpalMediaFormat::DestroyContents()
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   delete m_info;
 }
 
 
 PObject * OpalMediaFormat::Clone() const
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   return new OpalMediaFormat(*this);
 }
 
 
 PObject::Comparison OpalMediaFormat::Compare(const PObject & obj) const
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   PAssert(PIsDescendant(&obj, OpalMediaFormat), PInvalidCast);
   const OpalMediaFormat & other = (const OpalMediaFormat &)obj;
   if (m_info == NULL)
@@ -793,7 +804,7 @@ PObject::Comparison OpalMediaFormat::Compare(const PObject & obj) const
 
 void OpalMediaFormat::PrintOn(ostream & strm) const
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   if (m_info != NULL)
     strm << *m_info;
 }
@@ -801,7 +812,7 @@ void OpalMediaFormat::PrintOn(ostream & strm) const
 
 void OpalMediaFormat::ReadFrom(istream & strm)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   char fmt[100];
   strm >> fmt;
   operator=(fmt);
@@ -810,7 +821,7 @@ void OpalMediaFormat::ReadFrom(istream & strm)
 
 bool OpalMediaFormat::ToNormalisedOptions()
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   MakeUnique();
   return m_info != NULL && m_info->ToNormalisedOptions();
 }
@@ -818,20 +829,74 @@ bool OpalMediaFormat::ToNormalisedOptions()
 
 bool OpalMediaFormat::ToCustomisedOptions()
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   MakeUnique();
   return m_info != NULL && m_info->ToCustomisedOptions();
 }
 
 
+bool OpalMediaFormat::Update(const OpalMediaFormat & mediaFormat)
+{
+  if (!mediaFormat.IsValid())
+    return true;
+
+  PWaitAndSignal m(m_mutex);
+  MakeUnique();
+
+  if (*this != mediaFormat)
+    return Merge(mediaFormat);
+
+  if (!IsValid() || !Merge(mediaFormat))
+    *this = mediaFormat; //Must have different EqualMerge options, just copy it
+  else if (GetPayloadType() != mediaFormat.GetPayloadType()) {
+    PTRACE(4, "MediaFormat\tChanging payload type from " << GetPayloadType()
+           << " to " << mediaFormat.GetPayloadType() << " in " << *this);
+    SetPayloadType(mediaFormat.GetPayloadType());
+  }
+
+  return true;
+}
+
+
 bool OpalMediaFormat::Merge(const OpalMediaFormat & mediaFormat)
 {
-  PWaitAndSignal m(_mutex);
+  PWaitAndSignal m(m_mutex);
   MakeUnique();
   return m_info != NULL && mediaFormat.m_info != NULL && m_info->Merge(*mediaFormat.m_info);
 }
 
 
+bool OpalMediaFormat::ValidateMerge(const OpalMediaFormat & mediaFormat) const
+{
+  PWaitAndSignal m(m_mutex);
+  return m_info != NULL && mediaFormat.m_info != NULL && m_info->ValidateMerge(*mediaFormat.m_info);
+}
+
+
+OpalMediaFormatList OpalMediaFormat::GetMatchingRegisteredMediaFormats(RTP_DataFrame::PayloadTypes rtpPayloadType,
+                                                                       const unsigned clockRate,
+                                                                       const char * rtpEncodingName,
+                                                                       const char * protocol)
+{
+  OpalMediaFormatList matches;
+
+  PWaitAndSignal mutex(GetMediaFormatsListMutex());
+
+  const OpalMediaFormatList & registeredFormats = GetMediaFormatsList();
+  OpalMediaFormatList::const_iterator format = registeredFormats.FindFormat(rtpPayloadType, clockRate, rtpEncodingName, protocol);
+  while (format != registeredFormats.end()) {
+    matches += *format;
+
+    if (++format == registeredFormats.end())
+      break;
+
+    format = registeredFormats.FindFormat(rtpPayloadType, clockRate, rtpEncodingName, protocol, format);
+  }
+
+  return matches;
+}
+
+
 OpalMediaFormatList OpalMediaFormat::GetAllRegisteredMediaFormats()
 {
   OpalMediaFormatList copy;
@@ -891,7 +956,8 @@ OpalMediaFormatInternal::OpalMediaFormatInternal(const char * fullName,
   if (nj)
     AddOption(new OpalMediaOptionBoolean(OpalMediaFormat::NeedsJitterOption(), true, OpalMediaOption::OrMerge, true));
 
-  AddOption(new OpalMediaOptionUnsigned(OpalMediaFormat::MaxBitRateOption(), true, OpalMediaOption::MinMerge, bw, 100));
+  if (bw > 0)
+    AddOption(new OpalMediaOptionUnsigned(OpalMediaFormat::MaxBitRateOption(), true, OpalMediaOption::MinMerge, bw, 100));
 
   if (fs > 0)
     AddOption(new OpalMediaOptionUnsigned(OpalMediaFormat::MaxFrameSizeOption(), true, OpalMediaOption::NoMerge, fs));
@@ -981,6 +1047,29 @@ bool OpalMediaFormatInternal::Merge(const OpalMediaFormatInternal & mediaFormat)
 }
 
 
+bool OpalMediaFormatInternal::ValidateMerge(const OpalMediaFormatInternal & mediaFormat) const
+{
+  PWaitAndSignal m1(media_format_mutex);
+  PWaitAndSignal m2(mediaFormat.media_format_mutex);
+
+  for (PINDEX i = 0; i < options.GetSize(); i++) {
+    OpalMediaOption & opt = options[i];
+    PString name = opt.GetName();
+    OpalMediaOption * option = mediaFormat.FindOption(opt.GetName());
+    if (option == NULL) {
+      PTRACE_IF(2, formatName == mediaFormat.formatName, "MediaFormat\tValidate: unmatched option " << opt.GetName());
+    }
+    else {
+      PAssert(option->GetName() == opt.GetName(), "find returned bad name");
+      if (!opt.ValidateMerge(*option))
+        return false;
+    }
+  }
+
+  return true;
+}
+
+
 bool OpalMediaFormatInternal::ToNormalisedOptions()
 {
   return true;
@@ -1389,7 +1478,7 @@ OpalAudioFormatInternal::OpalAudioFormatInternal(const char * fullName,
   if (rxFrames > 0)
     AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::RxFramesPerPacketOption(), false, OpalMediaOption::NoMerge,  rxFrames, 1, maxFrames));
   if (txFrames > 0)
-    AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::TxFramesPerPacketOption(), false, OpalMediaOption::NoMerge, txFrames, 1, maxFrames));
+    AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::TxFramesPerPacketOption(), false, OpalMediaOption::AlwaysMerge, txFrames, 1, maxFrames));
 
   AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::MaxFramesPerPacketOption(), true,  OpalMediaOption::NoMerge,  maxFrames));
   AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::ChannelsOption(),           false, OpalMediaOption::NoMerge,  1, 1, 5));
@@ -1430,6 +1519,8 @@ const PString & OpalVideoFormat::TemporalSpatialTradeOffOption()  { static PStri
 const PString & OpalVideoFormat::TxKeyFramePeriodOption()         { static PString s = PLUGINCODEC_OPTION_TX_KEY_FRAME_PERIOD;       return s; }
 const PString & OpalVideoFormat::RateControlEnableOption()        { static PString s = "Rate Control Enable";                        return s; }
 const PString & OpalVideoFormat::RateControllerOption()           { static PString s = "Rate Controller";                            return s; }
+const PString & OpalVideoFormat::ContentRoleOption()              { static PString s = "Content Role";                               return s; }
+const PString & OpalVideoFormat::ContentRoleMaskOption()          { static PString s = "Content Role Mask";                          return s; }
 
 OpalVideoFormat::OpalVideoFormat(const char * fullName,
                                  RTP_DataFrame::PayloadTypes rtpPayloadType,
@@ -1477,11 +1568,27 @@ OpalVideoFormatInternal::OpalVideoFormatInternal(const char * fullName,
   AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::MaxRxFrameWidthOption(),          false, OpalMediaOption::MinMerge,    maxFrameWidth,               16,  32767));
   AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::MaxRxFrameHeightOption(),         false, OpalMediaOption::MinMerge,    maxFrameHeight,              16,  32767));
   AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::TargetBitRateOption(),            false, OpalMediaOption::AlwaysMerge, maxBitRate,                  1000      ));
-  AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::TxKeyFramePeriodOption(),         false, OpalMediaOption::AlwaysMerge, 125,                         0,   1000));
+  AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::TxKeyFramePeriodOption(),         false, OpalMediaOption::AlwaysMerge, 125,                         0,    1000));
   AddOption(new OpalMediaOptionBoolean (OpalVideoFormat::RateControlEnableOption(),        false, OpalMediaOption::NoMerge,     false                                  ));
-  AddOption(new OpalMediaOptionUnsigned(OpalMediaFormat::MaxTxPacketSizeOption(),          false, OpalMediaOption::NoMerge,     2048,                        100,  2048));
+  AddOption(new OpalMediaOptionUnsigned(OpalMediaFormat::MaxTxPacketSizeOption(),          false, OpalMediaOption::NoMerge,     8192,                        100,  8192));
   AddOption(new OpalMediaOptionString  (OpalVideoFormat::RateControllerOption(),           false                                                                       ));
 
+
+  static const char * const RoleEnumerations[OpalVideoFormat::eNumRoles] = {
+    "No Role",
+    "Presentation",
+    "Main",
+    "Speaker",
+    "Sign Language"
+  };
+  AddOption(new OpalMediaOptionEnum(OpalVideoFormat::ContentRoleOption(), false,
+                                    RoleEnumerations, PARRAYSIZE(RoleEnumerations),
+                                    OpalMediaOption::NoMerge));
+
+  AddOption(new OpalMediaOptionUnsigned(OpalVideoFormat::ContentRoleMaskOption(),
+                                        false, OpalMediaOption::AndMerge,
+                                        0, 0, OpalVideoFormat::ContentRoleMask));
+
   // For video the max bit rate and frame rate is adjustable by user
   FindOption(OpalVideoFormat::MaxBitRateOption())->SetReadOnly(false);
   FindOption(OpalVideoFormat::FrameTimeOption())->SetReadOnly(false);
@@ -1510,6 +1617,16 @@ bool OpalVideoFormatInternal::Merge(const OpalMediaFormatInternal & mediaFormat)
   return true;
 }
 
+
+void OpalMediaFormat::AdjustVideoArgs(PVideoDevice::OpenArgs & args) const
+{
+  args.width = GetOptionInteger(OpalVideoFormat::FrameWidthOption(), PVideoFrameInfo::QCIFWidth);
+  args.height = GetOptionInteger(OpalVideoFormat::FrameHeightOption(), PVideoFrameInfo::QCIFHeight);
+  unsigned maxRate = GetClockRate()/GetFrameTime();
+  if (args.rate > maxRate)
+    args.rate = maxRate;
+}
+
 #endif // OPAL_VIDEO
 
 ///////////////////////////////////////////////////////////////////////////////
@@ -1565,10 +1682,12 @@ OpalMediaFormatList & OpalMediaFormatList::operator-=(const OpalMediaFormatList
 
 void OpalMediaFormatList::Remove(const PStringArray & maskList)
 {
+  if (maskList.IsEmpty())
+    return;
+
   PTRACE(4,"MediaFormat\tRemoving codecs " << setfill(',') << maskList);
 
-  PINDEX i;
-  for (i = 0; i < maskList.GetSize(); i++) {
+  for (PINDEX i = 0; i < maskList.GetSize(); i++) {
     OpalMediaFormatList::const_iterator fmt;
     while ((fmt = FindFormat(maskList[i])) != end())
       erase(fmt);
@@ -1576,13 +1695,12 @@ void OpalMediaFormatList::Remove(const PStringArray & maskList)
 }
 
 
-OpalMediaFormatList::const_iterator OpalMediaFormatList::FindFormat(RTP_DataFrame::PayloadTypes pt, unsigned clockRate, const char * name, const char * protocol) const
+OpalMediaFormatList::const_iterator OpalMediaFormatList::FindFormat(RTP_DataFrame::PayloadTypes pt, unsigned clockRate, const char * name, const char * protocol, const_iterator iter) const
 {
-  OpalMediaFormatList::const_iterator format;
-
   // First look for a matching encoding name
   if (name != NULL && *name != '\0') {
-    for (format = begin(); format != end(); ++format) {
+    OpalMediaFormatList::const_iterator format ((iter == const_iterator()) ? begin() : iter);
+    for (; format != end(); ++format) {
       // If encoding name matches exactly, then use it regardless of payload code.
       const char * otherName = format->GetEncodingName();
       if (otherName != NULL && strcasecmp(otherName, name) == 0 &&
@@ -1597,7 +1715,8 @@ OpalMediaFormatList::const_iterator OpalMediaFormatList::FindFormat(RTP_DataFram
   // someone to override a standard payload type with another encoding name, so
   // have to search all formats by name before trying by number.
   if (pt < RTP_DataFrame::DynamicBase) {
-    for (format = begin(); format != end(); ++format) {
+    OpalMediaFormatList::const_iterator format ((iter == const_iterator()) ? begin() : iter);
+    for (; format != end(); ++format) {
       if (format->GetPayloadType() == pt &&
           (clockRate == 0    || clockRate == format->GetClockRate()) && // if have clock rate, clock rate must match
           (protocol  == NULL || format->IsValidForProtocol(protocol))) // if protocol is specified, must be valid for the protocol
@@ -1716,7 +1835,7 @@ bool OpalMediaFormatList::HasType(const OpalMediaType & type, bool mustBeTranspo
 {
   OpalMediaFormatList::const_iterator format;
   for (format = begin(); format != end(); ++format) {
-    if (format->GetMediaType() == type && (mustBeTransportable || format->IsTransportable()))
+    if (format->GetMediaType() == type && (!mustBeTransportable || format->IsTransportable()))
       return true;
   }
 
