@@ -23,9 +23,9 @@
  *
  * Contributor(s): ________________________________________.
  *
- * $Revision: 23149 $
+ * $Revision: 24156 $
  * $Author: rjongbloed $
- * $Date: 2009-07-28 19:37:30 -0500 (Tue, 28 Jul 2009) $
+ * $Date: 2010-03-23 23:12:09 -0500 (Tue, 23 Mar 2010) $
  */
 
 #include <ptlib.h>
@@ -49,7 +49,7 @@
 #include <rtp/rtp.h>
 #include <opal/transports.h>
 #include <opal/rtpconn.h>
-#include <opal/endpoint.h>
+#include <opal/rtpep.h>
 #include <opal/call.h>
 
 #define MAX_PAYLOAD_TYPE_MISMATCHES 10
@@ -114,13 +114,10 @@ PBoolean OpalMediaStream::UpdateMediaFormat(const OpalMediaFormat & newMediaForm
   if (mediaPatch != NULL && !fromPatch)
     return mediaPatch->UpdateMediaFormat(newMediaFormat);
 
-  if (mediaFormat != newMediaFormat)
-    return mediaFormat.Merge(newMediaFormat);
-
-  mediaFormat = newMediaFormat;
+  if (!mediaFormat.Update(newMediaFormat))
+    return false;
 
   PTRACE(4, "Media\tMedia format updated on " << *this);
-
   return true;
 }
 
@@ -247,9 +244,16 @@ inline static unsigned CalculateTimestamp(PINDEX size, const OpalMediaFormat & m
 
 PBoolean OpalMediaStream::ReadPacket(RTP_DataFrame & packet)
 {
+  if (!isOpen)
+    return false;
+
   unsigned oldTimestamp = timestamp;
 
-  PAssert(defaultDataSize <= (packet.GetSize() - packet.GetHeaderSize()), "default data size too large for packet");
+  if (defaultDataSize < (packet.GetSize() - RTP_DataFrame::MinHeaderSize)) {
+    stringstream str;
+    str << "Media stream buffer " << (packet.GetSize() - RTP_DataFrame::MinHeaderSize) << " too small for media packet " << defaultDataSize;
+    PAssertAlways(str.str().c_str());
+  }
 
   PINDEX lastReadCount;
   if (!ReadData(packet.GetPayloadPtr(), defaultDataSize, lastReadCount))
@@ -275,6 +279,9 @@ PBoolean OpalMediaStream::ReadPacket(RTP_DataFrame & packet)
 
 PBoolean OpalMediaStream::WritePacket(RTP_DataFrame & packet)
 {
+  if (!isOpen)
+    return false;
+
   if (paused)
     packet.SetPayloadSize(0);
 
@@ -384,11 +391,14 @@ PBoolean OpalMediaStream::WriteData(const BYTE * buffer, PINDEX length, PINDEX &
 
 PBoolean OpalMediaStream::PushPacket(RTP_DataFrame & packet)
 {
-  if(mediaPatch == NULL) {
-    return false;
+  OpalMediaPatch * patch = NULL;
+  if (LockReadOnly()) {
+    patch = mediaPatch;
+    UnlockReadOnly();
   }
-	
-  return mediaPatch->PushFrame(packet);
+
+  // OpalMediaPatch::PushFrame() might block, do outside of mutex
+  return patch != NULL && patch->PushFrame(packet);
 }
 
 
@@ -429,11 +439,38 @@ void OpalMediaStream::SetPaused(bool p)
 
 PBoolean OpalMediaStream::SetPatch(OpalMediaPatch * patch)
 {
+#if PTRACING
+  if (PTrace::CanTrace(4) && (patch != NULL || mediaPatch != NULL)) {
+    ostream & trace = PTrace::Begin(4, __FILE__, __LINE__);
+    if (patch == NULL)
+      trace << "Removing patch " << *mediaPatch;
+    else if (mediaPatch == NULL)
+      trace << "Adding patch " << *patch;
+    else
+      trace << "Overwriting patch " << *mediaPatch << " with " << *patch;
+    trace << " on stream " << *this << PTrace::End;
+  }
+#endif
+
   PSafeLockReadWrite safeLock(*this);
   if (!safeLock.IsLocked())
     return false;
 
+  if (mediaPatch == NULL) {
+    mediaPatch = patch;
+    return true;
+  }
+
+  OpalMediaPatch * oldPatch = mediaPatch;
   mediaPatch = patch;
+
+  if (IsSink())
+    oldPatch->RemoveSink(this);
+  else {
+    oldPatch->Close();
+    connection.GetEndPoint().GetManager().DestroyMediaPatch(oldPatch);
+  }
+
   return true;
 }
 
@@ -441,10 +478,7 @@ PBoolean OpalMediaStream::SetPatch(OpalMediaPatch * patch)
 void OpalMediaStream::AddFilter(const PNotifier & Filter, const OpalMediaFormat & Stage)
 {
   PSafeLockReadWrite safeLock(*this);
-  if (!safeLock.IsLocked())
-    return;
-
-  if (mediaPatch != NULL)
+  if (safeLock.IsLocked() && mediaPatch != NULL)
     mediaPatch->AddFilter(Filter, Stage);
 }
 
@@ -452,12 +486,7 @@ void OpalMediaStream::AddFilter(const PNotifier & Filter, const OpalMediaFormat
 PBoolean OpalMediaStream::RemoveFilter(const PNotifier & Filter, const OpalMediaFormat & Stage)
 {
   PSafeLockReadWrite safeLock(*this);
-  if (!safeLock.IsLocked())
-    return false;
-
-  if (mediaPatch != NULL) return mediaPatch->RemoveFilter(Filter, Stage);
-
-  return false;
+  return safeLock.IsLocked() && mediaPatch != NULL && mediaPatch->RemoveFilter(Filter, Stage);
 }
 
 
@@ -465,10 +494,7 @@ PBoolean OpalMediaStream::RemoveFilter(const PNotifier & Filter, const OpalMedia
 void OpalMediaStream::GetStatistics(OpalMediaStatistics & statistics, bool fromPatch) const
 {
   PSafeLockReadOnly safeLock(*this);
-  if (!safeLock.IsLocked())
-    return;
-
-  if (mediaPatch != NULL && !fromPatch)
+  if (safeLock.IsLocked() && mediaPatch != NULL && !fromPatch)
     mediaPatch->GetStatistics(statistics, IsSink());
 }
 #endif
@@ -480,14 +506,14 @@ void OpalMediaStream::RemovePatch(OpalMediaPatch * /*patch*/ )
 }
 
 
-void OpalMediaStream::OnPatchStart() 
+void OpalMediaStream::OnStartMediaPatch() 
 { 
-  connection.OnMediaPatchStart(sessionID, isSource);
+  connection.OnStartMediaPatch(*mediaPatch);
 }
 
-void OpalMediaStream::OnPatchStop() 
+void OpalMediaStream::OnStopMediaPatch(OpalMediaPatch & patch)
 { 
-  connection.OnMediaPatchStop(sessionID, isSource);
+  connection.OnStopMediaPatch(patch);
 }
 
 
@@ -530,6 +556,21 @@ OpalNullMediaStream::OpalNullMediaStream(OpalConnection & conn,
   : OpalMediaStream(conn, mediaFormat, sessionID, isSource)
   , OpalMediaStreamPacing(mediaFormat)
   , m_isSynchronous(isSyncronous)
+  , m_requiresPatchThread(isSyncronous)
+{
+}
+
+
+OpalNullMediaStream::OpalNullMediaStream(OpalConnection & conn,
+                                         const OpalMediaFormat & mediaFormat,
+                                         unsigned sessionID,
+                                         bool isSource,
+                                         bool usePacingDelay,
+                                         bool requiresPatchThread)
+  : OpalMediaStream(conn, mediaFormat, sessionID, isSource)
+  , OpalMediaStreamPacing(mediaFormat)
+  , m_isSynchronous(usePacingDelay)
+  , m_requiresPatchThread(requiresPatchThread)
 {
 }
 
@@ -542,7 +583,8 @@ PBoolean OpalNullMediaStream::ReadData(BYTE * buffer, PINDEX size, PINDEX & leng
   memset(buffer, 0, size);
   length = size;
 
-  Pace(true, size, marker);
+  if (m_isSynchronous)
+    Pace(true, size, marker);
   return true;
 }
 
@@ -554,14 +596,15 @@ PBoolean OpalNullMediaStream::WriteData(const BYTE * /*buffer*/, PINDEX length,
 
   written = length != 0 ? length : defaultDataSize;
 
-  Pace(false, written, marker);
+  if (m_isSynchronous)
+    Pace(false, written, marker);
   return true;
 }
 
 
 PBoolean OpalNullMediaStream::RequiresPatchThread() const
 {
-  return m_isSynchronous;
+  return m_requiresPatchThread;
 }
 
 
@@ -591,7 +634,7 @@ OpalRTPMediaStream::OpalRTPMediaStream(OpalRTPConnection & conn,
      practical UDP packet size. This means we have a buffer that can accept
      whatever the RTP sender throws at us. For sink, we set it to the
      maximum size based on MTU (or other criteria). */
-  defaultDataSize = isSource ? 2048 : conn.GetMaxRtpPayloadSize();
+  defaultDataSize = isSource ? conn.GetEndPoint().GetManager().GetMaxRtpPacketSize() : conn.GetMaxRtpPayloadSize();
 }
 
 
@@ -627,6 +670,27 @@ PBoolean OpalRTPMediaStream::Close()
 }
 
 
+void OpalRTPMediaStream::SetPaused(bool pause)
+{
+  if (paused == pause)
+    return;
+
+  OpalMediaStream::SetPaused(pause);
+
+  // If coming out of pause, reopen the RTP session, even though it is probably
+  // still open, to make sure any pending error/statistic conditions are reset.
+  if (!paused)
+    rtpSession.Reopen(IsSource());
+
+  if (IsSource()) {
+    if (pause)
+      rtpSession.SetJitterBufferSize(0, 0);
+    else
+      EnableJitterBuffer();
+  }
+}
+
+
 PBoolean OpalRTPMediaStream::ReadPacket(RTP_DataFrame & packet)
 {
   if (IsSink()) {
@@ -634,8 +698,12 @@ PBoolean OpalRTPMediaStream::ReadPacket(RTP_DataFrame & packet)
     return false;
   }
 
-  if (!rtpSession.ReadBufferedData(packet))
-    return false;
+  if (paused)
+    packet.SetPayloadSize(0);
+  else {
+    if (!rtpSession.ReadBufferedData(packet))
+      return false;
+  }
 
   timestamp = packet.GetTimestamp();
   return true;
@@ -644,9 +712,6 @@ PBoolean OpalRTPMediaStream::ReadPacket(RTP_DataFrame & packet)
 
 PBoolean OpalRTPMediaStream::WritePacket(RTP_DataFrame & packet)
 {
-  if (paused)
-    packet.SetPayloadSize(0);
-  
   if (IsSource()) {
     PTRACE(1, "Media\tTried to write to source media stream");
     return false;
@@ -654,9 +719,10 @@ PBoolean OpalRTPMediaStream::WritePacket(RTP_DataFrame & packet)
 
   timestamp = packet.GetTimestamp();
 
-  if (packet.GetPayloadSize() == 0)
+  if (paused || packet.GetPayloadSize() == 0)
     return true;
 
+  packet.SetPayloadType(mediaFormat.GetPayloadType());
   return rtpSession.WriteData(packet);
 }
 
@@ -670,16 +736,41 @@ PBoolean OpalRTPMediaStream::SetDataSize(PINDEX PTRACE_PARAM(dataSize), PINDEX /
 
 PBoolean OpalRTPMediaStream::IsSynchronous() const
 {
-  return IsSource() && !mediaFormat.NeedsJitterBuffer();
+  return false;
+}
+
+
+PBoolean OpalRTPMediaStream::RequiresPatchThread() const
+{
+  return !dynamic_cast<OpalRTPEndPoint &>(connection.GetEndPoint()).CheckForLocalRTP(*this);
 }
 
 
 void OpalRTPMediaStream::EnableJitterBuffer() const
 {
-  if (mediaFormat.NeedsJitterBuffer())
-    rtpSession.SetJitterBufferSize(minAudioJitterDelay*mediaFormat.GetTimeUnits(),
-                                   maxAudioJitterDelay*mediaFormat.GetTimeUnits(),
-				                       mediaFormat.GetTimeUnits());
+  unsigned minJitter, maxJitter;
+  if (mediaFormat.NeedsJitterBuffer()) {
+    minJitter = minAudioJitterDelay*mediaFormat.GetTimeUnits();
+    maxJitter = maxAudioJitterDelay*mediaFormat.GetTimeUnits();
+  }
+  else
+    minJitter = maxJitter = 0;
+
+  rtpSession.SetJitterBufferSize(minJitter, maxJitter,
+                                 mediaFormat.GetTimeUnits(),
+                                 connection.GetEndPoint().GetManager().GetMaxRtpPacketSize());
+}
+
+
+PBoolean OpalRTPMediaStream::SetPatch(OpalMediaPatch * patch)
+{
+  if (!isOpen || IsSink())
+    return OpalMediaStream::SetPatch(patch);
+
+  rtpSession.Close(true);
+  bool ok = OpalMediaStream::SetPatch(patch);
+  rtpSession.Reopen(true);
+  return ok;
 }
 
 
@@ -933,11 +1024,13 @@ OpalAudioMediaStream::OpalAudioMediaStream(OpalConnection & conn,
                                            unsigned sessionID,
                                            PBoolean isSource,
                                            PINDEX buffers,
+                                           unsigned bufferTime,
                                            PSoundChannel * channel,
                                            PBoolean autoDel)
   : OpalRawMediaStream(conn, mediaFormat, sessionID, isSource, channel, autoDel)
+  , m_soundChannelBuffers(buffers)
+  , m_soundChannelBufferTime(bufferTime)
 {
-  soundChannelBuffers = buffers;
 }
 
 
@@ -946,6 +1039,7 @@ OpalAudioMediaStream::OpalAudioMediaStream(OpalConnection & conn,
                                            unsigned sessionID,
                                            PBoolean isSource,
                                            PINDEX buffers,
+                                           unsigned bufferTime,
                                            const PString & deviceName)
   : OpalRawMediaStream(conn, mediaFormat, sessionID, isSource,
                        PSoundChannel::CreateOpenedChannel(PString::Empty(),
@@ -954,8 +1048,9 @@ OpalAudioMediaStream::OpalAudioMediaStream(OpalConnection & conn,
                                                                    : PSoundChannel::Player,
                                                           1, mediaFormat.GetClockRate(), 16),
                        true)
+  , m_soundChannelBuffers(buffers)
+  , m_soundChannelBufferTime(bufferTime)
 {
-  soundChannelBuffers = buffers;
 }
 
 
@@ -967,19 +1062,24 @@ PBoolean OpalAudioMediaStream::SetDataSize(PINDEX dataSize, PINDEX frameTime)
   const unsigned BufferTimeMilliseconds = 10;
   PINDEX bufferSize = BufferTimeMilliseconds*mediaFormat.GetClockRate()/1000*sizeof(short);
 
+  // Calculate sound buffers from global system settings
+  PINDEX soundChannelBuffers = (m_soundChannelBufferTime+BufferTimeMilliseconds-1)/BufferTimeMilliseconds;
+  if (soundChannelBuffers < m_soundChannelBuffers)
+    soundChannelBuffers = m_soundChannelBuffers;
+
   // Quantise dataSize up to multiple of the frame time
   PINDEX frameSize = frameTime*sizeof(short);
   dataSize = (dataSize+frameSize-1)/frameSize * frameSize;
 
   // Calculte buffers needed for the data we require and increase buffers if required
   PINDEX dataBuffersNeeded = (dataSize+bufferSize-1)/bufferSize;
-  if (dataBuffersNeeded < soundChannelBuffers)
-    dataBuffersNeeded = soundChannelBuffers;
+  if (soundChannelBuffers < dataBuffersNeeded)
+    soundChannelBuffers = dataBuffersNeeded;
 
   PTRACE(3, "Media\tAudio " << (IsSource() ? "source" : "sink") << " data size set to "
-         << dataSize << ", buffer size set to " << bufferSize << " and " << dataBuffersNeeded << " buffers.");
+         << dataSize << ", buffer size set to " << bufferSize << " and " << soundChannelBuffers << " buffers.");
   return OpalMediaStream::SetDataSize(dataSize, frameTime) &&
-         ((PSoundChannel *)m_channel)->SetBuffers(bufferSize, dataBuffersNeeded);
+         ((PSoundChannel *)m_channel)->SetBuffers(bufferSize, soundChannelBuffers);
 }
 
 
@@ -1126,15 +1226,16 @@ PBoolean OpalVideoMediaStream::ReadData(BYTE * data, PINDEX size, PINDEX & lengt
     return false;
 
   PTimeInterval currentGrabTime = PTimer::Tick();
-  timestamp += ((lastGrabTime - currentGrabTime)*1000/OpalMediaFormat::VideoClockRate).GetInterval();
+  timestamp += (int)((currentGrabTime - lastGrabTime).GetMilliSeconds()*OpalMediaFormat::VideoClockRate/1000);
   lastGrabTime = currentGrabTime;
 
-  marker = true;
-  length = bytesReturned + sizeof(PluginCodec_Video_FrameHeader);
-
-  if ((flags & PluginCodec_ReturnCoderRequestIFrame) != 0) {
+  if ((flags & PluginCodec_ReturnCoderRequestIFrame) != 0)
     ExecuteCommand(OpalVideoUpdatePicture());
-  }
+
+  marker = true;
+  length = bytesReturned;
+  if (length > 0)
+    length += sizeof(PluginCodec_Video_FrameHeader);
 
   if (outputDevice == NULL)
     return true;
@@ -1209,6 +1310,11 @@ OpalUDPMediaStream::OpalUDPMediaStream(OpalConnection & conn,
 {
 }
 
+OpalUDPMediaStream::~OpalUDPMediaStream()
+{
+  Close();
+}
+
 
 PBoolean OpalUDPMediaStream::ReadPacket(RTP_DataFrame & Packet)
 {
