@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 24346 $
+ * $Revision: 24130 $
  * $Author: rjongbloed $
- * $Date: 2010-05-17 20:29:28 -0500 (Mon, 17 May 2010) $
+ * $Date: 2010-03-17 02:06:06 -0500 (Wed, 17 Mar 2010) $
  */
 
 #include <ptlib.h>
@@ -37,12 +37,17 @@
 
 #include <opal.h>
 #include <opal/manager.h>
+
+#if OPAL_HAS_PCSS
 #include <opal/pcss.h>
+#endif
+
 #include <opal/localep.h>
 #include <h323/h323ep.h>
 #include <sip/sipep.h>
 #include <iax2/iax2ep.h>
 #include <lids/lidep.h>
+#include <t38/t38proto.h>
 #include <opal/ivr.h>
 
 #include <queue>
@@ -89,6 +94,7 @@ class OpalMessageBuffer
 
 
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
 
 class OpalPCSSEndPoint_C : public OpalPCSSEndPoint
 {
@@ -102,6 +108,7 @@ class OpalPCSSEndPoint_C : public OpalPCSSEndPoint
     OpalManager_C & m_manager;
 };
 
+#endif
 #endif // OPAL_PTLIB_AUDIO
 
 
@@ -116,10 +123,12 @@ class OpalLocalEndPoint_C : public OpalLocalEndPoint
     virtual bool OnWriteMediaFrame(const OpalLocalConnection &, const OpalMediaStream &, RTP_DataFrame & frame);
     virtual bool OnReadMediaData(const OpalLocalConnection &, const OpalMediaStream &, void *, PINDEX, PINDEX &);
     virtual bool OnWriteMediaData(const OpalLocalConnection &, const OpalMediaStream &, const void *, PINDEX, PINDEX &);
+    virtual bool IsSynchronous() const;
 
     OpalMediaDataFunction m_mediaReadData;
     OpalMediaDataFunction m_mediaWriteData;
     OpalMediaDataType     m_mediaDataHeader;
+    OpalMediaTiming       m_mediaTiming;
 
   private:
     OpalManager_C & m_manager;
@@ -158,9 +167,12 @@ class OpalManager_C : public OpalManager
     OpalManager_C(unsigned version)
       : localEP(NULL)
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
       , pcssEP(NULL)
+#endif
 #endif
       , m_apiVersion(version)
+      , m_manualAlerting(false)
       , m_messagesAvailable(0, INT_MAX)
     {
     }
@@ -185,11 +197,14 @@ class OpalManager_C : public OpalManager
     virtual void OnProceeding(OpalConnection & conenction);
     virtual void OnClearedCall(OpalCall & call);
 
+    bool IsManualAlerting() const { return m_manualAlerting; }
+
   private:
     void HandleSetGeneral    (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleSetProtocol   (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleRegistration  (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleSetUpCall     (const OpalMessage & message, OpalMessageBuffer & response);
+    void HandleAlerting      (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleAnswerCall    (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleUserInput     (const OpalMessage & message, OpalMessageBuffer & response);
     void HandleClearCall     (const OpalMessage & message, OpalMessageBuffer & response);
@@ -207,10 +222,13 @@ class OpalManager_C : public OpalManager
 
     OpalLocalEndPoint_C * localEP;
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
     OpalPCSSEndPoint_C  * pcssEP;
+#endif
 #endif
 
     unsigned                  m_apiVersion;
+    bool                      m_manualAlerting;
     std::queue<OpalMessage *> m_messageQueue;
     PMutex                    m_messageMutex;
     PSemaphore                m_messagesAvailable;
@@ -367,16 +385,16 @@ OpalLocalEndPoint_C::OpalLocalEndPoint_C(OpalManager_C & mgr)
   , m_mediaReadData(NULL)
   , m_mediaWriteData(NULL)
   , m_mediaDataHeader(OpalMediaDataPayloadOnly)
+  , m_mediaTiming(OpalMediaTimingSynchronous)
   , m_manager(mgr)
 {
+  m_deferredAlerting = mgr.IsManualAlerting();
 }
 
 
-bool OpalLocalEndPoint_C::OnOutgoingCall(const OpalLocalConnection & connection)
+static void SetOutgoingCallInfo(OpalMessageBuffer & message, const OpalConnection & connection)
 {
-  PTRACE(4, "OpalC\tOnOutgoingCall " << connection);
   const OpalCall & call = connection.GetCall();
-  OpalMessageBuffer message(OpalIndAlerting);
   SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_partyA, call.GetPartyA());
   SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_partyB, call.GetPartyB());
   SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_callToken, call.GetToken());
@@ -384,12 +402,19 @@ bool OpalLocalEndPoint_C::OnOutgoingCall(const OpalLocalConnection & connection)
             " token=\"" << message->m_param.m_callSetUp.m_callToken << "\""
             " A=\""     << message->m_param.m_callSetUp.m_partyA << "\""
             " B=\""     << message->m_param.m_callSetUp.m_partyB << '"');
+}
+
+
+bool OpalLocalEndPoint_C::OnOutgoingCall(const OpalLocalConnection & connection)
+{
+  OpalMessageBuffer message(OpalIndAlerting);
+  SetOutgoingCallInfo(message, connection);
   m_manager.PostMessage(message);
   return true;
 }
 
 
-static void SetIncomingCall(OpalMessageBuffer & message, const OpalConnection & connection)
+static void SetIncomingCallInfo(OpalMessageBuffer & message, const OpalConnection & connection)
 {
   PSafePtr<OpalConnection> network = connection.GetOtherPartyConnection();
   PAssert(network != NULL, PLogicError); // Should not happen!
@@ -428,9 +453,8 @@ static void SetIncomingCall(OpalMessageBuffer & message, const OpalConnection &
 
 bool OpalLocalEndPoint_C::OnIncomingCall(OpalLocalConnection & connection)
 {
-  PTRACE(4, "OpalC\tOnIncomingCall " << connection);
   OpalMessageBuffer message(OpalIndIncomingCall);
-  SetIncomingCall(message, connection);
+  SetIncomingCallInfo(message, connection);
   m_manager.PostMessage(message);
   return true;
 }
@@ -532,22 +556,29 @@ bool OpalLocalEndPoint_C::OnWriteMediaData(const OpalLocalConnection & connectio
 }
 
 
+bool OpalLocalEndPoint_C::IsSynchronous() const
+{
+  return m_mediaTiming == OpalMediaTimingSynchronous;
+}
+
+
 ///////////////////////////////////////
 
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
 
 OpalPCSSEndPoint_C::OpalPCSSEndPoint_C(OpalManager_C & mgr)
   : OpalPCSSEndPoint(mgr)
   , m_manager(mgr)
 {
+  m_deferredAlerting = mgr.IsManualAlerting();
 }
 
 
 PBoolean OpalPCSSEndPoint_C::OnShowIncoming(const OpalPCSSConnection & connection)
 {
-  PTRACE(4, "OpalC\tOnShowIncoming " << connection);
   OpalMessageBuffer message(OpalIndIncomingCall);
-  SetIncomingCall(message, connection);
+  SetIncomingCallInfo(message, connection);
   m_manager.PostMessage(message);
   return true;
 }
@@ -555,20 +586,13 @@ PBoolean OpalPCSSEndPoint_C::OnShowIncoming(const OpalPCSSConnection & connectio
 
 PBoolean OpalPCSSEndPoint_C::OnShowOutgoing(const OpalPCSSConnection & connection)
 {
-  PTRACE(4, "OpalC\tOnShowOutgoing " << connection);
-  const OpalCall & call = connection.GetCall();
   OpalMessageBuffer message(OpalIndAlerting);
-  SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_partyA, call.GetPartyA());
-  SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_partyB, call.GetPartyB());
-  SET_MESSAGE_STRING(message, m_param.m_callSetUp.m_callToken, call.GetToken());
-  PTRACE(4, "OpalC API\tOnShowOutgoing:"
-            " token=\"" << message->m_param.m_callSetUp.m_callToken << "\""
-            " A=\""     << message->m_param.m_callSetUp.m_partyA << "\""
-            " B=\""     << message->m_param.m_callSetUp.m_partyB << '"');
+  SetOutgoingCallInfo(message, connection);
   m_manager.PostMessage(message);
   return true;
 }
 
+#endif
 #endif // OPAL_PTLIB_AUDIO
 
 
@@ -730,6 +754,20 @@ bool OpalManager_C::Initialise(const PCaselessString & options)
   }
 #endif
 
+#if OPAL_FAX
+  PINDEX faxPos = options.Find("fax");
+  if (faxPos < defUserPos) {
+    defUser = "fax:";
+    defUserPos = faxPos;
+  }
+
+  PINDEX t38Pos = options.Find("t38");
+  if (t38Pos < defProtoPos) {
+    defUser = "t38:";
+    defProtoPos = t38Pos;
+  }
+#endif
+
   PINDEX pcPos = options.Find("pc");
   if (pcPos < defUserPos) {
     defUser = "pc:*";
@@ -782,11 +820,24 @@ bool OpalManager_C::Initialise(const PCaselessString & options)
   }
 #endif
 
+#if OPAL_FAX
+  if (faxPos != P_MAX_INDEX || t38Pos != P_MAX_INDEX) {
+    new OpalFaxEndPoint(*this);
+
+    if (faxPos != P_MAX_INDEX)
+      AddRouteEntry("fax:.*=" + defProto + ":<da>");
+    if (t38Pos != P_MAX_INDEX)
+      AddRouteEntry("t38:.*=" + defUser + ":<da>");
+  }
+#endif
+
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
   if (pcPos != P_MAX_INDEX) {
     pcssEP = new OpalPCSSEndPoint_C(*this);
     AddRouteEntry("pc:.*=" + defProto + ":<da>");
   }
+#endif
 #endif
 
   if (localPos != P_MAX_INDEX) {
@@ -889,6 +940,7 @@ OpalMessage * OpalManager_C::SendMessage(const OpalMessage * message)
 void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuffer & response)
 {
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
   if (pcssEP != NULL) {
     SET_MESSAGE_STRING(response, m_param.m_general.m_audioRecordDevice, pcssEP->GetSoundChannelRecordDevice());
     if (!IsNullString(command.m_param.m_general.m_audioRecordDevice))
@@ -898,6 +950,7 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
     if (!IsNullString(command.m_param.m_general.m_audioPlayerDevice))
       pcssEP->SetSoundChannelPlayDevice(command.m_param.m_general.m_audioPlayerDevice);
   }
+#endif
 #endif // OPAL_PTLIB_AUDIO
 
 #if OPAL_VIDEO
@@ -935,23 +988,39 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
   if (!IsNullString(command.m_param.m_general.m_mediaMask))
     SetMediaFormatMask(PString(command.m_param.m_general.m_mediaMask).Lines());
 
-  strm = "audio";
-#if OPAL_VIDEO
-  if (CanAutoStartReceiveVideo())
-    strm << " video";
-  if (!IsNullString(command.m_param.m_general.m_autoRxMedia))
-    SetAutoStartReceiveVideo(strstr(command.m_param.m_general.m_autoRxMedia, "video") != NULL);
-#endif
-  SET_MESSAGE_STRING(response, m_param.m_general.m_autoRxMedia, strm);
+  OpalMediaTypeFactory::KeyList_T allMediaTypes = OpalMediaType::GetList();
 
-  strm = "audio";
-#if OPAL_VIDEO
-  if (CanAutoStartTransmitVideo())
-    strm << " video";
-  if (!IsNullString(command.m_param.m_general.m_autoTxMedia))
-    SetAutoStartTransmitVideo(strstr(command.m_param.m_general.m_autoTxMedia, "video") != NULL);
-#endif
-  SET_MESSAGE_STRING(response, m_param.m_general.m_autoTxMedia, strm);
+  for (OpalMediaType::AutoStartMode autoStart = OpalMediaType::Receive; autoStart < OpalMediaType::ReceiveTransmit; ++autoStart) {
+    strm.MakeEmpty();
+
+    OpalMediaTypeFactory::KeyList_T::iterator iterMediaType;
+    for (iterMediaType = allMediaTypes.begin(); iterMediaType != allMediaTypes.end(); ++iterMediaType) {
+      OpalMediaTypeDefinition * definition = OpalMediaType::GetDefinition(*iterMediaType);
+      if ((definition->GetAutoStart()&autoStart) != 0) {
+        if (!strm.IsEmpty())
+          strm << ' ';
+        strm << *iterMediaType;
+        definition->SetAutoStart(autoStart, false);
+      }
+    }
+
+    PString autoXxMedia;
+    if (autoStart == OpalMediaType::Receive) {
+      SET_MESSAGE_STRING(response, m_param.m_general.m_autoRxMedia, strm);
+      autoXxMedia = command.m_param.m_general.m_autoRxMedia;
+    }
+    else {
+      SET_MESSAGE_STRING(response, m_param.m_general.m_autoTxMedia, strm);
+      autoXxMedia = command.m_param.m_general.m_autoTxMedia;
+    }
+
+    PStringArray enabledMediaTypes = autoXxMedia.Tokenise(" \t\n", false);
+    for (PINDEX i = 0; i < enabledMediaTypes.GetSize(); ++i) {
+      OpalMediaTypeDefinition * definition = OpalMediaType::GetDefinition(enabledMediaTypes[0]);
+      if (definition != NULL)
+        definition->SetAutoStart(autoStart, true);
+    }
+  }
 
   SET_MESSAGE_STRING(response, m_param.m_general.m_natRouter, GetTranslationHost());
   if (!IsNullString(command.m_param.m_general.m_natRouter)) {
@@ -1020,21 +1089,25 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
     silenceDetectParams.m_adaptivePeriod = command.m_param.m_general.m_silenceAdaptPeriod;
   SetSilenceDetectParams(silenceDetectParams);
 
+#if OPAL_AEC
   OpalEchoCanceler::Params echoCancelParams = GetEchoCancelParams();
   response->m_param.m_general.m_echoCancellation = (OpalEchoCancelMode)(echoCancelParams.m_mode+1);
   if (command.m_param.m_general.m_echoCancellation != 0)
     echoCancelParams.m_mode = (OpalEchoCanceler::Mode)(command.m_param.m_general.m_echoCancellation-1);
   SetEchoCancelParams(echoCancelParams);
+#endif
 
   if (m_apiVersion < 3)
     return;
 
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
   if (pcssEP != NULL) {
     response->m_param.m_general.m_audioBuffers = pcssEP->GetSoundChannelBufferDepth();
     if (command.m_param.m_general.m_audioBuffers != 0)
       pcssEP->SetSoundChannelBufferDepth(command.m_param.m_general.m_audioBuffers);
   }
+#endif
 #endif
 
   if (m_apiVersion < 5)
@@ -1052,6 +1125,12 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
     response->m_param.m_general.m_mediaDataHeader = localEP->m_mediaDataHeader;
     if (command.m_param.m_general.m_mediaDataHeader != 0)
       localEP->m_mediaDataHeader = command.m_param.m_general.m_mediaDataHeader;
+
+    if (m_apiVersion >= 20) {
+      response->m_param.m_general.m_mediaTiming = localEP->m_mediaTiming;
+      if (command.m_param.m_general.m_mediaTiming != 0)
+        localEP->m_mediaTiming = command.m_param.m_general.m_mediaTiming;
+    }
   }
 
   if (m_apiVersion < 8)
@@ -1065,6 +1144,20 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
   if (m_apiVersion < 14)
     return;
 
+  OpalMediaFormatList allCodecs = OpalMediaFormat::GetAllRegisteredMediaFormats();
+
+  PStringStream mediaOptions;
+  for (OpalMediaFormatList::iterator itMediaFormat = allCodecs.begin(); itMediaFormat != allCodecs.end(); ++itMediaFormat) {
+    if (itMediaFormat->IsTransportable()) {
+      mediaOptions << *itMediaFormat << ":Media Type#" << itMediaFormat->GetMediaType() << '\n';
+      for (PINDEX i = 0; i < itMediaFormat->GetOptionCount(); ++i) {
+        const OpalMediaOption & option = itMediaFormat->GetOption(i);
+        mediaOptions << *itMediaFormat << ':' << option.GetName() << (option.IsReadOnly() ? '#' : '=') << option << '\n';
+      }
+    }
+  }
+  SET_MESSAGE_STRING(response, m_param.m_general.m_mediaOptions, mediaOptions);
+
   PStringArray options = PString(command.m_param.m_general.m_mediaOptions).Lines();
   for (PINDEX i = 0; i < options.GetSize(); ++i) {
     PString optionSpec = options[i];
@@ -1081,7 +1174,6 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
       OpalMediaType mediaType = mediaName.ToLower();
       if (OpalMediaTypeFactory::CreateInstance(mediaType) != NULL) {
         // Known media type name, change all codecs of that type
-        OpalMediaFormatList allCodecs = OpalMediaFormat::GetAllRegisteredMediaFormats();
         for (OpalMediaFormatList::iterator it = allCodecs.begin(); it != allCodecs.end(); ++it) {
           if (it->GetMediaType() == mediaType) {
             if (it->SetOptionValue(optionName, optionValue)) {
@@ -1115,6 +1207,26 @@ void OpalManager_C::HandleSetGeneral(const OpalMessage & command, OpalMessageBuf
       }
     }
   }
+
+  if (m_apiVersion < 17)
+    return;
+
+#if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
+  if (pcssEP != NULL) {
+    response->m_param.m_general.m_audioBufferTime = pcssEP->GetSoundChannelBufferTime();
+    if (command.m_param.m_general.m_audioBufferTime != 0)
+      pcssEP->SetSoundChannelBufferTime(command.m_param.m_general.m_audioBufferTime);
+  }
+#endif
+#endif
+
+  if (m_apiVersion < 19)
+    return;
+
+  response->m_param.m_general.m_audioBufferTime = m_manualAlerting ? 2 : 1;
+  if (command.m_param.m_general.m_manualAlerting != 0)
+    m_manualAlerting = command.m_param.m_general.m_manualAlerting != 1;
 }
 
 
@@ -1333,9 +1445,11 @@ void OpalManager_C::HandleSetUpCall(const OpalMessage & command, OpalMessageBuff
   PString partyA = command.m_param.m_callSetUp.m_partyA;
   if (partyA.IsEmpty()) {
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
     if (pcssEP != NULL)
       partyA = "pc:*";
     else
+#endif
 #endif
     if (localEP != NULL)
       partyA = "local:*";
@@ -1364,6 +1478,38 @@ void OpalManager_C::HandleSetUpCall(const OpalMessage & command, OpalMessageBuff
 }
 
 
+void OpalManager_C::HandleAlerting(const OpalMessage & command, OpalMessageBuffer & response)
+{
+  if (IsNullString(command.m_param.m_callToken)) {
+    response.SetError("No call token provided.");
+    return;
+  }
+
+  if (
+#if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
+      pcssEP == NULL &&
+#endif
+#endif
+      localEP == NULL) {
+    response.SetError("Can only control alerting from PC.");
+    return;
+  }
+
+#if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
+  if (pcssEP != NULL && pcssEP->AlertingIncomingCall(command.m_param.m_callToken))
+    return;
+#endif
+#endif
+
+  if (localEP != NULL && localEP->AlertingIncomingCall(command.m_param.m_callToken))
+    return;
+
+  response.SetError("No call found by the token provided.");
+}
+
+
 void OpalManager_C::HandleAnswerCall(const OpalMessage & command, OpalMessageBuffer & response)
 {
   if (IsNullString(command.m_param.m_callToken)) {
@@ -1373,7 +1519,9 @@ void OpalManager_C::HandleAnswerCall(const OpalMessage & command, OpalMessageBuf
 
   if (
 #if OPAL_PTLIB_AUDIO
+#if OPAL_HAS_PCSS
       pcssEP == NULL &&
+#endif
 #endif
       localEP == NULL) {
     response.SetError("Can only answer calls to PC.");
@@ -1381,8 +1529,10 @@ void OpalManager_C::HandleAnswerCall(const OpalMessage & command, OpalMessageBuf
   }
 
 #if OPAL_PTLIB_AUDIO
-  if (pcssEP != NULL && pcssEP->AcceptIncomingConnection(command.m_param.m_callToken))
+#if OPAL_HAS_PCSS
+  if (pcssEP != NULL && pcssEP->AcceptIncomingCall(command.m_param.m_callToken))
     return;
+#endif
 #endif
 
   if (localEP != NULL && localEP->AcceptIncomingCall(command.m_param.m_callToken))
@@ -1447,7 +1597,7 @@ void OpalManager_C::HandleClearCall(const OpalMessage & command, OpalMessageBuff
   }
   else {
     callToken = command.m_param.m_clearCall.m_callToken;
-    reason = (OpalConnection::CallEndReason)command.m_param.m_clearCall.m_reason;
+    reason.code = (OpalConnection::CallEndReasonCodes)command.m_param.m_clearCall.m_reason;
   }
 
   if (IsNullString(callToken)) {
@@ -1515,10 +1665,7 @@ void OpalManager_C::HandleTransferCall(const OpalMessage & command, OpalMessageB
     }
   }
 
-  if (connection->GetPhase() < OpalConnection::ConnectedPhase)
-    connection->ForwardCall(command.m_param.m_callSetUp.m_partyB);
-  else
-    connection->TransferConnection(command.m_param.m_callSetUp.m_partyB);
+  connection->TransferConnection(command.m_param.m_callSetUp.m_partyB);
 }
 
 
@@ -1596,6 +1743,7 @@ void OpalManager_C::HandleMediaStream(const OpalMessage & command, OpalMessageBu
 
 void OpalManager_C::HandleStartRecording(const OpalMessage & command, OpalMessageBuffer & response)
 {
+#if OPAL_HAS_MIXER
   PSafePtr<OpalCall> call;
   if (!FindCall(command.m_param.m_userInput.m_callToken, response, call))
     return;
@@ -1606,19 +1754,34 @@ void OpalManager_C::HandleStartRecording(const OpalMessage & command, OpalMessag
     return;
   }
 
-  if (!call->StartRecording(command.m_param.m_recording.m_file,
-                            command.m_param.m_recording.m_channels < 2))
+  OpalRecordManager::Options options;
+  options.m_stereo = command.m_param.m_recording.m_channels == 2;
+  if (m_apiVersion >= 21) {
+    options.m_audioFormat = command.m_param.m_recording.m_audioFormat;
+#if OPAL_VIDEO
+    options.m_videoFormat = command.m_param.m_recording.m_videoFormat;
+    options.m_videoWidth  = command.m_param.m_recording.m_videoWidth;
+    options.m_videoHeight = command.m_param.m_recording.m_videoHeight;
+    options.m_videoRate   = command.m_param.m_recording.m_videoRate;
+    options.m_videoMixing = (OpalRecordManager::VideoMode)command.m_param.m_recording.m_videoMixing;
+#endif
+  }
+
+  if (!call->StartRecording(command.m_param.m_recording.m_file, options))
+#endif
     response.SetError("Could not start recording for call.");
 }
 
 
 void OpalManager_C::HandleStopRecording(const OpalMessage & command, OpalMessageBuffer & response)
 {
+#if OPAL_HAS_MIXER
   PSafePtr<OpalCall> call;
   if (!FindCall(command.m_param.m_userInput.m_callToken, response, call))
     return;
 
   call->StopRecording();
+#endif
 }
 
 
@@ -1764,49 +1927,9 @@ void OpalManager_C::OnClearedCall(OpalCall & call)
   OpalMessageBuffer message(OpalIndCallCleared);
   SET_MESSAGE_STRING(message, m_param.m_callCleared.m_callToken, call.GetToken());
 
-  static const char * const CallEndReasonStrings[] = {
-    "Local party cleared call",
-    "Local party did not accept call",
-    "Local party declined to answer call",
-    "Remote party cleared call",
-    "Remote party refused call",
-    "Remote party did not answer in required time",
-    "Remote party stopped calling",
-    "Call failed due to a transport error",
-    "Connection to remote failed",
-    "Gatekeeper has cleared call",
-    "Call failed as could not find user",
-    "Call failed due to insufficient bandwidth",
-    "Call failed as could not find common media capabilities",
-    "Call was forwarded",
-    "Call failed security check",
-    "Local party busy",
-    "Local party congested",
-    "Remote party busy",
-    "Remote switch congested",
-    "Remote party could but be reached",
-    "Remote party application is not running",
-    "Remote party host is off line",
-    "Remote system failed temporarily",
-    "Call cleared with unmapped Q.931 cause code",
-    "Call cleared due to an enforced duration limit",
-    "Call cleared due to invalid conference ID",
-    "Call cleared due to missing dial tone",
-    "Call cleared due to missing ringback tone",
-    "Call cleared because the line is out of service",
-    "Call cleared because another call is answered"
-  };
 
   PStringStream str;
-  OpalConnection::CallEndReason reason = call.GetCallEndReason();
-  str << (unsigned)reason << ": ";
-
-  if (reason < PARRAYSIZE(CallEndReasonStrings))
-    str << CallEndReasonStrings[reason];
-  else if ((reason&OpalConnection::EndedWithQ931Code) != 0)
-    str << "Call cleared with Q.931 code " << (reason >> 8);
-  else
-    str << "Call cleared with unknown cause code.";
+  str << (unsigned)call.GetCallEndReason() << ": " << call.GetCallEndReasonText();
 
   SET_MESSAGE_STRING(message, m_param.m_callCleared.m_reason, str);
   PTRACE(4, "OpalC API\tOnClearedCall:"
@@ -1824,7 +1947,11 @@ extern "C" {
 
   OpalHandle OPAL_EXPORT OpalInitialise(unsigned * version, const char * options)
   {
-    PCaselessString optionsString = IsNullString(options) ? "pcss h323 sip iax2 pots pstn ivr" : options;
+    PCaselessString optionsString = IsNullString(options) ?
+#if OPAL_HAS_PCSS
+            "pcss "
+#endif
+            "h323 sip iax2 pots pstn fax t38 ivr" : options;
 
     unsigned callerVersion = 1;
     if (version != NULL) {
@@ -1869,4 +1996,354 @@ extern "C" {
 }; // extern "C"
 
 
+///////////////////////////////////////////////////////////////////////////////
+
+OpalContext::OpalContext()
+  : m_handle(NULL)
+{
+}
+
+
+OpalContext::~OpalContext()
+{
+  ShutDown();
+}
+
+
+unsigned OpalContext::Initialise(const char * options, unsigned version)
+{
+  ShutDown();
+
+  m_handle = OpalInitialise(&version, options);
+  return m_handle != NULL ? version : 0;
+}
+
+
+void OpalContext::ShutDown()
+{
+  if (m_handle != NULL) {
+    OpalShutDown(m_handle);
+    m_handle = NULL;
+  }
+}
+
+
+bool OpalContext::GetMessage(OpalMessagePtr & message, unsigned timeout)
+{
+  if (m_handle == NULL) {
+    message.SetType(OpalIndCommandError);
+    message.m_message->m_param.m_commandError = "Uninitialised OPAL context.";
+    return false;
+  }
+
+  message.m_message = OpalGetMessage(m_handle, timeout);
+  if (message.m_message != NULL)
+    return true;
+
+  message.SetType(OpalIndCommandError);
+  message.m_message->m_param.m_commandError = "Timeout getting message.";
+  return false;
+}
+
+
+bool OpalContext::SendMessage(const OpalMessagePtr & message, OpalMessagePtr & response)
+{
+  if (m_handle == NULL) {
+    response.SetType(OpalIndCommandError);
+    response.m_message->m_param.m_commandError = "Uninitialised OPAL context.";
+    return false;
+  }
+
+  response.m_message = OpalSendMessage(m_handle, message.m_message);
+  if (response.m_message != NULL)
+    return response.GetType() != OpalIndCommandError;
+
+  response.SetType(OpalIndCommandError);
+  response.m_message->m_param.m_commandError = "Invalid message.";
+  return false;
+}
+
+
+bool OpalContext::SetUpCall(OpalMessagePtr & response,
+                            const char * partyB,
+                            const char * partyA,
+                            const char * alertingType)
+{
+  OpalMessagePtr message(OpalCmdSetUpCall);
+  OpalParamSetUpCall * param = message.GetCallSetUp();
+  param->m_partyA = partyA;
+  param->m_partyB = partyB;
+  param->m_alertingType = alertingType;
+  return SendMessage(message, response);
+}
+
+
+bool OpalContext::AnswerCall(const char * callToken)
+{
+  OpalMessagePtr message(OpalCmdAnswerCall), response;
+  message.SetCallToken(callToken);
+  return SendMessage(message, response);
+}
+
+
+bool OpalContext::ClearCall(const char * callToken, OpalCallEndReason reason)
+{
+  OpalMessagePtr message(OpalCmdClearCall), response;
+  OpalParamCallCleared * param = message.GetClearCall();
+  param->m_callToken = callToken;
+  param->m_reason = reason;
+  return SendMessage(message, response);
+}
+
+
+bool OpalContext::SendUserInput(const char * callToken, const char * userInput, unsigned duration)
+{
+  OpalMessagePtr message(OpalCmdClearCall), response;
+  OpalParamUserInput * param = message.GetUserInput();
+  param->m_callToken = callToken;
+  param->m_userInput = userInput;
+  param->m_duration = duration;
+  return SendMessage(message, response);
+}
+
+
+OpalMessagePtr::OpalMessagePtr(OpalMessageType type)
+  : m_message(NULL)
+{
+  SetType(type);
+}
+
+
+OpalMessagePtr::~OpalMessagePtr()
+{
+  OpalFreeMessage(m_message);
+}
+
+
+OpalMessageType OpalMessagePtr::GetType() const
+{
+  return m_message->m_type;
+}
+
+
+void OpalMessagePtr::SetType(OpalMessageType type)
+{
+  OpalFreeMessage(m_message);
+
+  m_message = (OpalMessage *)malloc(sizeof(OpalMessage)); // Use malloc to be compatible with OpalFreeMessage
+  memset(m_message, 0, sizeof(OpalMessage));
+  m_message->m_type = type;
+}
+
+
+const char * OpalMessagePtr::GetCommandError() const
+{
+  return m_message->m_type == OpalIndCommandError ? m_message->m_param.m_commandError : NULL;
+}
+
+
+const char * OpalMessagePtr::GetCallToken() const
+{
+  switch (m_message->m_type) {
+    case OpalCmdAnswerCall :
+    case OpalCmdHoldCall :
+    case OpalCmdRetrieveCall :
+    case OpalCmdStopRecording :
+    case OpalCmdAlerting :
+      return m_message->m_param.m_callToken;
+
+    case OpalCmdSetUpCall :
+    case OpalIndProceeding :
+    case OpalIndAlerting :
+    case OpalIndEstablished :
+      return m_message->m_param.m_callSetUp.m_callToken;
+
+    case OpalIndIncomingCall :
+      return m_message->m_param.m_incomingCall.m_callToken;
+
+    case OpalIndMediaStream :
+    case OpalCmdMediaStream :
+      return m_message->m_param.m_mediaStream.m_callToken;
+
+    case OpalCmdSetUserData :
+      return m_message->m_param.m_setUserData.m_callToken;
+
+    case OpalIndUserInput :
+      return m_message->m_param.m_userInput.m_callToken;
+
+    case OpalCmdStartRecording :
+      return m_message->m_param.m_recording.m_callToken;
+
+    case OpalIndCallCleared :
+      return m_message->m_param.m_callCleared.m_callToken;
+
+    case OpalCmdClearCall :
+      return m_message->m_param.m_clearCall.m_callToken;
+
+    default :
+      return NULL;
+  }
+}
+
+
+void OpalMessagePtr::SetCallToken(const char * callToken)
+{
+  switch (m_message->m_type) {
+    case OpalCmdAnswerCall :
+    case OpalCmdHoldCall :
+    case OpalCmdRetrieveCall :
+    case OpalCmdStopRecording :
+    case OpalCmdAlerting :
+      m_message->m_param.m_callToken = callToken;
+      break;
+
+    case OpalCmdSetUpCall :
+    case OpalIndProceeding :
+    case OpalIndAlerting :
+    case OpalIndEstablished :
+      m_message->m_param.m_callSetUp.m_callToken = callToken;
+      break;
+
+    case OpalIndIncomingCall :
+      m_message->m_param.m_incomingCall.m_callToken = callToken;
+      break;
+
+    case OpalIndMediaStream :
+    case OpalCmdMediaStream :
+      m_message->m_param.m_mediaStream.m_callToken = callToken;
+      break;
+
+    case OpalCmdSetUserData :
+      m_message->m_param.m_setUserData.m_callToken = callToken;
+      break;
+
+    case OpalIndUserInput :
+      m_message->m_param.m_userInput.m_callToken = callToken;
+      break;
+
+    case OpalCmdStartRecording :
+      m_message->m_param.m_recording.m_callToken = callToken;
+      break;
+
+    case OpalIndCallCleared :
+      m_message->m_param.m_callCleared.m_callToken = callToken;
+      break;
+
+    case OpalCmdClearCall :
+      m_message->m_param.m_clearCall.m_callToken = callToken;
+      break;
+
+    default :
+      break;
+  }
+}
+
+
+OpalParamGeneral * OpalMessagePtr::GetGeneralParams() const
+{
+  return m_message->m_type == OpalCmdSetGeneralParameters ? &m_message->m_param.m_general : NULL;
+}
+
+
+OpalParamProtocol * OpalMessagePtr::GetProtocolParams() const
+{
+  return m_message->m_type == OpalCmdSetProtocolParameters ? &m_message->m_param.m_protocol : NULL;
+}
+
+
+OpalParamRegistration * OpalMessagePtr::GetRegistrationInfo() const
+{
+  return m_message->m_type == OpalCmdRegistration ? &m_message->m_param.m_registrationInfo : NULL;
+}
+
+
+OpalStatusRegistration * OpalMessagePtr::GetRegistrationStatus() const
+{
+  return m_message->m_type == OpalIndRegistration ? &m_message->m_param.m_registrationStatus : NULL;
+}
+
+
+OpalParamSetUpCall * OpalMessagePtr::GetCallSetUp() const
+{
+  switch (m_message->m_type) {
+    case OpalCmdSetUpCall :
+    case OpalIndProceeding :
+    case OpalIndAlerting :
+    case OpalIndEstablished :
+      return &m_message->m_param.m_callSetUp;
+
+    default :
+      return NULL;
+  }
+}
+
+
+OpalStatusIncomingCall * OpalMessagePtr::GetIncomingCall() const
+{
+  return m_message->m_type == OpalIndIncomingCall ? &m_message->m_param.m_incomingCall : NULL;
+}
+
+
+OpalStatusUserInput * OpalMessagePtr::GetUserInput() const
+{
+  switch (m_message->m_type) {
+    case OpalIndUserInput :
+    case OpalCmdUserInput :
+      return &m_message->m_param.m_userInput;
+
+    default :
+      return NULL;
+  }
+}
+
+
+OpalStatusMessageWaiting * OpalMessagePtr::GetMessageWaiting() const
+{
+  return m_message->m_type == OpalIndMessageWaiting ? &m_message->m_param.m_messageWaiting : NULL;
+}
+
+
+OpalStatusLineAppearance * OpalMessagePtr::GetLineAppearance() const
+{
+  return m_message->m_type == OpalIndLineAppearance ? &m_message->m_param.m_lineAppearance : NULL;
+}
+
+
+OpalStatusCallCleared * OpalMessagePtr::GetCallCleared() const
+{
+  return m_message->m_type == OpalIndCallCleared ? &m_message->m_param.m_callCleared : NULL;
+}
+
+
+OpalParamCallCleared * OpalMessagePtr::GetClearCall() const
+{
+  return m_message->m_type == OpalCmdClearCall ? &m_message->m_param.m_clearCall : NULL;
+}
+
+
+OpalStatusMediaStream * OpalMessagePtr::GetMediaStream() const
+{
+  switch (m_message->m_type) {
+    case OpalIndMediaStream :
+    case OpalCmdMediaStream :
+      return &m_message->m_param.m_mediaStream;
+
+    default :
+      return NULL;
+  }
+}
+
+
+OpalParamSetUserData * OpalMessagePtr::GetSetUserData() const
+{
+  return m_message->m_type == OpalCmdSetUserData ? &m_message->m_param.m_setUserData : NULL;
+}
+
+
+OpalParamRecording * OpalMessagePtr::GetRecording() const
+{
+  return m_message->m_type == OpalCmdStartRecording ? &m_message->m_param.m_recording : NULL;
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
